import json
from dataclasses import dataclass, field
from typing import Dict, List, Union

from vigorish.scrape.bbref_boxscores.models.boxscore_game_meta import BBRefBoxscoreMeta
from vigorish.scrape.bbref_boxscores.models.boxscore_team_data import BBRefBoxscoreTeamData
from vigorish.scrape.bbref_boxscores.models.half_inning import BBRefHalfInning
from vigorish.scrape.bbref_boxscores.models.umpire import BBRefUmpire
from vigorish.util.list_helpers import as_dict_list
from vigorish.util.string_helpers import get_brooks_team_id, validate_bbref_game_id


@dataclass
class BBRefBoxscore:
    """Batting and pitching statistics for a single MLB game."""

    boxscore_url: str
    bbref_game_id: str
    game_meta_info: BBRefBoxscoreMeta = None
    away_team_data: BBRefBoxscoreTeamData = None
    home_team_data: BBRefBoxscoreTeamData = None
    player_id_match_log: Dict[str, Union[str, int, List[str]]] = field(default_factory=dict)
    player_team_dict: Dict[str, str] = field(default_factory=dict)
    player_name_dict: Dict[str, str] = field(default_factory=dict)
    innings_list: List[BBRefHalfInning] = field(default_factory=list)
    umpires: List[BBRefUmpire] = field(default_factory=list)

    @property
    def bb_game_id(self):
        result = validate_bbref_game_id(self.bbref_game_id)
        game_date = result.value["game_date"]
        game_number = result.value["game_number"]
        away_team_id = get_brooks_team_id(self.away_team_data.team_id_br).lower()
        home_team_id = get_brooks_team_id(self.home_team_data.team_id_br).lower()
        return (
            f"gid_{game_date.year}_{game_date.month:02d}_{game_date.day:02d}_"
            f"{away_team_id}mlb_{home_team_id}mlb_{game_number}"
        )

    @property
    def game_date(self):
        result = validate_bbref_game_id(self.bbref_game_id)
        return result.value["game_date"] if result.success else None

    @property
    def away_team_pitch_appearance_count(self):
        return len(self.away_team_data.pitching_stats)

    @property
    def home_team_pitch_appearance_count(self):
        return len(self.home_team_data.pitching_stats)

    @property
    def pitch_appearance_count(self):
        return self.away_team_pitch_appearance_count + self.home_team_pitch_appearance_count

    @property
    def away_team_pitch_count(self):
        return sum(int(pitch_stats.pitch_count) for pitch_stats in self.away_team_data.pitching_stats)

    @property
    def home_team_pitch_count(self):
        return sum(int(pitch_stats.pitch_count) for pitch_stats in self.home_team_data.pitching_stats)

    @property
    def pitch_count(self):
        return self.away_team_pitch_count + self.home_team_pitch_count

    def __repr__(self):
        return f"<BBRefBoxscore bbref_game_id={self.bbref_game_id}>"

    def as_dict(self):
        """Convert boxscore to a dictionary."""
        return {
            "__bbref_boxscore__": True,
            "boxscore_url": self.boxscore_url,
            "bbref_game_id": self.bbref_game_id,
            "game_meta_info": self.game_meta_info.as_dict(),
            "away_team_data": self.away_team_data.as_dict(),
            "home_team_data": self.home_team_data.as_dict(),
            "innings_list": as_dict_list(self.innings_list),
            "player_id_match_log": self.player_id_match_log,
            "umpires": as_dict_list(self.umpires),
            "player_team_dict": self.player_team_dict,
            "player_name_dict": self.player_name_dict,
        }

    def as_json(self):
        """Convert boxscore to JSON."""
        return json.dumps(self.as_dict(), indent=2, sort_keys=False)



from dataclasses import dataclass


@dataclass
class BBRefUmpire:
    """Name and field location of one umpire for a single game."""

    field_location: str = ""
    umpire_name: str = ""

    def as_dict(self):
        """Convert umpire name and field location to a dictionary."""
        return {
            "field_location": f"{self.field_location}",
            "umpire_name": f"{self.umpire_name}",
        }



from dataclasses import dataclass

from vigorish.enums import PlayByPlayEvent


@dataclass
class BBRefPlayByPlayMiscEvent:
    """Information describing a challenged play or other game event."""

    inning_id: str = ""
    inning_label: str = ""
    pbp_table_row_number: str = "0"
    description: str = ""

    @property
    def event_type(self):
        return PlayByPlayEvent.MISC

    def as_dict(self):
        return {
            "__bbref_pbp_misc_event__": True,
            "inning_id": self.inning_id,
            "inning_label": self.inning_label,
            "pbp_table_row_number": int(self.pbp_table_row_number),
            "description": self.description,
        }



from abc import ABC, abstractmethod
from pathlib import Path

from vigorish.util.numeric_helpers import validate_year_value
from vigorish.util.sys_helpers import validate_folder_path

YEAR_TOKEN = "{year}"
DATA_SET_TOKEN = "{data_set}"


class FolderPathSetting(ABC):
    _path_str: str

    def __init__(self, path, data_set):
        self._path_str = path
        self.data_set = data_set

    def __str__(self):
        return self._path_str

    @abstractmethod
    def resolve(self, year=None):  # pragma: no cover
        pass


class S3FolderPathSetting(FolderPathSetting):
    def resolve(self, year=None):
        path_str = self._path_str
        if path_str[-1] == "/":
            path_str = path_str[:-1]
        if YEAR_TOKEN in self._path_str:
            year = validate_year_value(year)
            path_str = path_str.replace(YEAR_TOKEN, str(year))
        if DATA_SET_TOKEN in self._path_str:
            path_str = path_str.replace(DATA_SET_TOKEN, self.data_set.name.lower())
        return path_str


class LocalFolderPathSetting(FolderPathSetting):
    def is_valid(self, year):
        absolute_path = self.resolve(year)
        return validate_folder_path(absolute_path)

    def resolve(self, year=None):
        path_str = self._path_str
        if YEAR_TOKEN in self._path_str:
            year = validate_year_value(year)
            path_str = path_str.replace(YEAR_TOKEN, str(year))
        if DATA_SET_TOKEN in self._path_str:
            path_str = path_str.replace(DATA_SET_TOKEN, self.data_set.name.lower())
        return str(Path(path_str).resolve())



"""Menu that allows the user to view data for games where all data has been scraped."""
import subprocess

from bullet import Input
from getch import pause

import vigorish.database as db
from vigorish.cli.components.prompts import select_game_prompt, user_options_prompt
from vigorish.cli.components.util import print_error, print_heading
from vigorish.cli.menu_item import MenuItem
from vigorish.cli.menu_items.view_game_data import ViewGameData
from vigorish.constants import EMOJIS, MENU_NUMBERS
from vigorish.util.result import Result
from vigorish.util.string_helpers import validate_bbref_game_id


class ViewGameDataMenu(MenuItem):
    def __init__(self, app):
        super().__init__(app)
        self.menu_item_text = "View Scraped Game Data"
        self.menu_item_emoji = EMOJIS.get("MICROSCOPE", "")
        self.exit_menu = False
        self.menu_option = None
        self.mlb_season = None
        self.team_id = None
        self.game_id = None

    def launch(self):
        while True:
            result = self.prompt_user_choose_method_to_select_game()
            if result.failure:
                break
            self.menu_option = result.value
            while True:
                if self.menu_option == "MANUAL":
                    result = self.enter_game_id()
                if self.menu_option == "SEASON":
                    result = self.select_season_game_id()
                if self.menu_option == "TEAM":
                    result = self.select_team_game_id()
                if result.failure:
                    break
        return Result.Ok()

    def prompt_user_choose_method_to_select_game(self):
        subprocess.run(["clear"])
        print_heading("Scraped Data Viewer - Please choose an option below", fg="bright_yellow")
        prompt = "How would you like to locate the game you wish to view?"
        choices = {
            f"{MENU_NUMBERS.get(1)}  Enter Game ID Manually": "MANUAL",
            f"{MENU_NUMBERS.get(2)}  Browse by Team": "TEAM",
            f"{MENU_NUMBERS.get(3)}  Browse by Season": "SEASON",
            f"{EMOJIS.get('BACK')} Return to Previous Menu": None,
        }
        return user_options_prompt(choices, prompt, clear_screen=False)

    def enter_game_id(self):
        subprocess.run(["clear"])
        print_heading("Scraped Data Viewer - Enter a BBRef Game ID", fg="bright_yellow")
        check_game_id = Input("Enter a BBRef Game ID: ").launch()
        try:
            result = validate_bbref_game_id(check_game_id)
        except ValueError:
            error = f"\n'{check_game_id}' is NOT a valid BBRef Game ID, please try again."
            print_error(error)
            pause(message="Press any key to continue...")
            return Result.Fail("")
        self.game_id = result.value["game_id"]
        game_date = result.value["game_date"]
        all_valid_game_ids = self.app.audit_report[game_date.year]["successful"]
        if self.game_id not in all_valid_game_ids:
            error = f"\nRequirements to show data for {check_game_id} have not been met"
            print_error(error)
            pause(message="Press any key to continue...")
            return Result.Fail("")
        view_game_data = ViewGameData(self.app, self.game_id)
        view_game_data.launch()
        return Result.Fail("")

    def select_season_game_id(self):
        while True:
            result = self.select_season_prompt()
            if result.failure:
                return result
            self.mlb_season = result.value
            game_ids = self.app.audit_report[self.mlb_season]["successful"]
            while True:
                subprocess.run(["clear"])
                heading = f"Scraped Data Viewer - Select Game (MLB Season: {self.mlb_season})"
                print_heading(heading, fg="bright_yellow")
                result = select_game_prompt(game_ids, use_numbers=False, clear_screen=False)
                if result.failure:
                    break
                self.game_id = result.value
                view_game_data = ViewGameData(self.app, self.game_id)
                view_game_data.launch()

    def select_team_game_id(self):
        while True:
            result = self.select_season_prompt()
            if result.failure:
                return result
            self.mlb_season = result.value
            while True:
                result = self.select_team_prompt()
                if result.failure:
                    self.team_id = None
                    break
                game_ids = result.value
                while True:
                    subprocess.run(["clear"])
                    heading = (
                        "Scraped Data Viewer - Select Game (MLB Season: {self.mlb_season}, db.Team: {self.team_id})"
                    )
                    print_heading(heading, fg="bright_yellow")
                    result = select_game_prompt(game_ids, use_numbers=False, clear_screen=False)
                    if result.failure:
                        break
                    self.game_id = result.value
                    view_game_data = ViewGameData(self.app, self.game_id)
                    view_game_data.launch()

    def select_season_prompt(self):
        subprocess.run(["clear"])
        print_heading("Scraped Data Viewer - Select MLB Season", fg="bright_yellow")
        prompt = "Select an MLB season from the list below:"
        choices = {
            f"{MENU_NUMBERS.get(num, str(num))}  {year}": year
            for num, (year, results) in enumerate(self.app.audit_report.items(), start=1)
            if results["successful"]
        }
        choices[f"{EMOJIS.get('BACK')} Return to Previous Menu"] = None
        return user_options_prompt(choices, prompt, clear_screen=False, first_item_is_exit_menu=True)

    def select_team_prompt(self):
        team_choices_dict = {
            t.team_id_br: t.name for t in db.Team.get_all_teams_for_season(self.db_session, year=self.mlb_season)
        }
        subprocess.run(["clear"])
        heading = f"Scraped Data Viewer - Select db.Team (MLB Season: {self.mlb_season})"
        print_heading(heading, fg="bright_yellow")
        prompt = "Select a team from the list below::"
        choices = {
            f"{EMOJIS.get('BLUE_DIAMOND')}  {name} ({team_id})": team_id for team_id, name in team_choices_dict.items()
        }
        choices[f"{EMOJIS.get('BACK')} Return to Previous Menu"] = None
        result = user_options_prompt(choices, prompt, clear_screen=False, first_item_is_exit_menu=True)
        if result.failure:
            return result
        self.team_id = result.value
        return self.get_team_game_ids()

    def get_team_game_ids(self):
        all_valid_game_ids = self.app.audit_report[self.mlb_season]["successful"]
        team_game_ids = db.GameScrapeStatus.get_all_bbref_game_ids_for_team(
            self.db_session, self.team_id, self.mlb_season
        )
        valid_team_game_ids = list(set(team_game_ids).intersection(set(all_valid_game_ids)))
        game_id_weights = self.get_game_id_weights(valid_team_game_ids)
        valid_team_game_ids.sort(key=lambda x: game_id_weights[x])
        return Result.Ok(valid_team_game_ids)

    def get_game_id_weights(self, game_ids):
        return {g: validate_bbref_game_id(g).value["game_date"] for g in game_ids}



"""Team statistics for a single season."""
from sqlalchemy import Column, ForeignKey, Integer, String
from sqlalchemy.orm import relationship

import vigorish.database as db
from vigorish.util.list_helpers import display_dict


class Team(db.Base):
    """Team statistics for a single season."""

    __tablename__ = "team"
    id = Column(Integer, primary_key=True)
    year = Column(Integer)
    league = Column(String)
    team_id = Column(String)
    franch_id = Column(String)
    division = Column(String)
    games = Column(Integer)
    games_at_home = Column(Integer)
    wins = Column(Integer)
    losses = Column(Integer)
    runs = Column(Integer)
    at_bats = Column(Integer)
    hits = Column(Integer)
    doubles = Column(Integer)
    triples = Column(Integer)
    homeruns = Column(Integer)
    base_on_balls = Column(Integer)
    strikeouts = Column(Integer)
    stolen_bases = Column(Integer)
    caught_stealing = Column(Integer)
    runs_against = Column(Integer)
    earned_runs = Column(Integer)
    saves = Column(Integer)
    ip_outs = Column(Integer)
    errors = Column(Integer)
    name = Column(String)
    park = Column(String)
    batting_park_factor = Column(Integer)
    pitching_park_factor = Column(Integer)
    team_id_br = Column(String)
    team_id_retro = Column(String)
    regular_season_id = Column(Integer, ForeignKey("season.id"))
    post_season_id = Column(Integer, ForeignKey("season.id"))

    regular_season = relationship("Season", foreign_keys=[regular_season_id])
    post_season = relationship("Season", foreign_keys=[post_season_id])
    players = relationship("Player", secondary="player_team", back_populates="teams")

    def __repr__(self):
        return f"<Team team_id={self.team_id}, year={self.year}>"

    def as_dict(self):
        return {c.name: getattr(self, c.name) for c in self.__table__.columns}

    def display(self):
        season_dict = self.as_dict()
        title = f"{self.name} {self.year}"
        display_dict(season_dict, title=title)

    @classmethod
    def find_by_team_id_and_year(cls, db_session, team_id_br, year):
        return db_session.query(cls).filter_by(team_id_br=team_id_br).filter_by(year=year).first()

    @classmethod
    def get_all_teams_for_season(cls, db_session, year):
        return db_session.query(cls).filter_by(year=year).all()

    @classmethod
    def get_team_id_map_for_year(cls, db_session, year):
        teams = cls.get_all_teams_for_season(db_session, year)
        return {t.team_id_br: t.id for t in teams}



import vigorish.database as db
from vigorish.enums import DataSet, ScrapeCondition
from vigorish.scrape.brooks_pitchfx.parse_html import parse_pitchfx_log
from vigorish.scrape.scrape_task import ScrapeTaskABC
from vigorish.status.update_status_brooks_pitchfx import update_pitch_appearance_status_records
from vigorish.util.dt_format_strings import DATE_ONLY_2
from vigorish.util.result import Result


class ScrapeBrooksPitchFx(ScrapeTaskABC):
    def __init__(self, app, db_job):
        self.data_set = DataSet.BROOKS_PITCHFX
        self.req_data_set = DataSet.BROOKS_PITCH_LOGS
        super().__init__(app, db_job)

    def check_prerequisites(self, game_date):
        brooks_pitch_logs = db.DateScrapeStatus.verify_all_brooks_pitch_logs_scraped_for_date(
            self.db_session, game_date
        )
        if brooks_pitch_logs:
            return Result.Ok()
        date_str = game_date.strftime(DATE_ONLY_2)
        error = (
            f"Brooks pitch logs for date {date_str} have not been scraped, unable to scrape "
            "Brooks pitchfx data until this has been done."
        )
        return Result.Fail(error)

    def check_current_status(self, game_date):
        if self.scrape_condition == ScrapeCondition.ALWAYS:
            return Result.Ok()
        scraped_brooks_pitchfx = db.DateScrapeStatus.verify_all_brooks_pitchfx_scraped_for_date(
            self.db_session, game_date
        )
        return Result.Ok() if not scraped_brooks_pitchfx else Result.Fail("skip")

    def parse_scraped_html(self):
        parsed = 0
        for game_date in self.date_range:
            pitch_logs_for_date = self.scraped_data.get_all_brooks_pitch_logs_for_date(game_date)
            if not pitch_logs_for_date:
                date_str = game_date.strftime(DATE_ONLY_2)
                error = f"Failed to retrieve {self.req_data_set} for date: {date_str}"
                return Result.Fail(error)
            for pitch_logs_for_game in pitch_logs_for_date:
                game_id = pitch_logs_for_game.bbref_game_id
                self.spinner.text = self.url_tracker.parse_html_report(parsed, game_id)
                for pitch_log in pitch_logs_for_game.pitch_logs:
                    if not pitch_log.parsed_all_info:
                        continue
                    if pitch_log.pitch_app_id not in self.url_tracker.parse_url_ids:
                        continue
                    html = self.url_tracker.get_html(pitch_log.pitch_app_id)
                    result = parse_pitchfx_log(html, pitch_log)
                    if result.failure:
                        return result
                    pitchfx_log = result.value
                    result = self.scraped_data.save_json(self.data_set, pitchfx_log)
                    if result.failure:
                        return Result.Fail(f"Error! {result.error} (ID: {pitch_log.pitch_app_id})")
                    result = self.update_status(pitchfx_log)
                    if result.failure:
                        return Result.Fail(f"Error! {result.error} (ID: {pitch_log.pitch_app_id})")
                    parsed += 1
                    self.spinner.text = self.url_tracker.parse_html_report(parsed, game_id)
                    self.db_session.commit()
        return Result.Ok()

    def parse_html(self, url_details):
        pass

    def update_status(self, parsed_data):
        return update_pitch_appearance_status_records(self.db_session, parsed_data)



from dataclasses import asdict, fields
from datetime import datetime
from typing import get_type_hints

from dacite import from_dict

from vigorish.util.dt_format_strings import DATE_ONLY


def get_field_types(data_class):
    field_names = [field.name for field in fields(data_class)]
    return {name: get_type_hints(data_class)[name] for name in field_names}


def serialize_dataclass_list_to_csv(data_class_objects, date_format):
    dataclass_dicts = [asdict(do) for do in data_class_objects]
    if not dataclass_dicts:
        return None
    col_names = [",".join(list(dataclass_dicts[0].keys()))]
    csv_rows = [dict_to_csv_row(d, date_format) for d in dataclass_dicts]
    return "\n".join((col_names + csv_rows))


def get_dataclass_list_from_csv(csv_filepath, data_class):
    if not csv_filepath.exists():
        return []
    csv_text = csv_filepath.read_text()
    if not csv_text:
        return []
    return deserialize_dataclass_list_from_csv(csv_text, data_class)


def deserialize_dataclass_list_from_csv(csv_text, data_class):
    csv_rows = csv_text.split("\n")
    col_names = [col.strip() for col in csv_rows.pop(0).split(",")]
    csv_rows = [row.split(",") for row in csv_rows]
    csv_dict_list = [dict(zip(col_names, row)) for row in csv_rows if row != [""]]
    return [from_dict(data_class=data_class, data=csv_dict) for csv_dict in csv_dict_list]


def serialize_db_object_to_csv(db_obj, dataclass, date_format=DATE_ONLY):
    csv_dict = {}
    for name, field in dataclass.__dataclass_fields__.items():
        value = getattr(db_obj, name, None)
        if field.type is bool:
            csv_dict[name] = bool(value)
        elif not value:
            csv_dict[name] = None
        elif field.type is int and not isinstance(value, int):
            csv_dict[name] = int(value)
        elif field.type is float and not isinstance(value, float):
            csv_dict[name] = float(value)
        else:
            csv_dict[name] = value
    return dict_to_csv_row(csv_dict, date_format)


def dict_to_csv_row(csv_dict, date_format):
    return ",".join(sanitize_value_for_csv(val, date_format) for val in csv_dict.values())


def sanitize_value_for_csv(val, date_format):
    return (
        val.replace(",", ";")
        if isinstance(val, str)
        else val.strftime(date_format).strip()
        if isinstance(val, datetime)
        else 0
        if (isinstance(val, bool) and not val)
        else 1
        if (isinstance(val, bool) and val)
        else ""
        if not val
        else str(val)
    )



"""Constant values that are referenced from multiple places."""
from vigorish.enums import DataSet

ENV_VAR_NAMES = [
    "AWS_ACCESS_KEY_ID",
    "AWS_SECRET_ACCESS_KEY",
    "AWS_DEFAULT_REGION",
    "CONFIG_FILE",
    "DATABASE_URL",
]

BR_BB_TEAM_ID_MAP = {
    "CHW": "CHA",
    "CHC": "CHN",
    "KCR": "KCA",
    "LAA": "ANA",
    "LAD": "LAN",
    "NYY": "NYA",
    "NYM": "NYN",
    "SDP": "SDN",
    "SFG": "SFN",
    "STL": "SLN",
    "TBR": "TBA",
    "WSN": "WAS",
}

BB_BR_TEAM_ID_MAP = {v: k for k, v in BR_BB_TEAM_ID_MAP.items()}

PACIFIC_TZ_TEAMS = ["ANA", "ARI", "LAD", "OAK", "SDP", "SEA", "SFG"]

DEFENSE_POSITIONS = [
    "P",
    "C",
    "1B",
    "2B",
    "3B",
    "SS",
    "RF",
    "CF",
    "LF",
    "DH",
    "PH",
    "PR",
]

PPB_PITCH_LOG_DICT = {
    "C": {
        "description": "Called strike",
        "pitch_counts": 1,
        "did_swing": False,
        "made_contact": False,
    },
    "S": {
        "description": "Swinging strike",
        "pitch_counts": 1,
        "did_swing": True,
        "made_contact": False,
    },
    "F": {"description": "Foul", "pitch_counts": 1, "did_swing": True, "made_contact": True},
    "B": {"description": "Ball", "pitch_counts": 1, "did_swing": False, "made_contact": False},
    "X": {
        "description": "Ball put into play by batter",
        "pitch_counts": 1,
        "did_swing": True,
        "made_contact": True,
    },
    "T": {"description": "Foul tip", "pitch_counts": 1, "did_swing": True, "made_contact": True},
    "K": {
        "description": "Strike (unknown type)",
        "pitch_counts": 1,
        "did_swing": False,
        "made_contact": False,
    },
    "I": {
        "description": "Intentional ball",
        "pitch_counts": 1,
        "did_swing": False,
        "made_contact": False,
    },
    "H": {
        "description": "Hit batter",
        "pitch_counts": 1,
        "did_swing": False,
        "made_contact": False,
    },
    "L": {"description": "Foul bunt", "pitch_counts": 1, "did_swing": True, "made_contact": True},
    "M": {
        "description": "Missed bunt attempt",
        "pitch_counts": 1,
        "did_swing": True,
        "made_contact": False,
    },
    "N": {
        "description": "No pitch (on balks and interference calls)",
        "pitch_counts": 0,
        "did_swing": False,
        "made_contact": False,
    },
    "O": {
        "description": "Foul tip on bunt",
        "pitch_counts": 1,
        "did_swing": True,
        "made_contact": True,
    },
    "P": {"description": "Pitchout", "pitch_counts": 1, "did_swing": False, "made_contact": False},
    "Q": {
        "description": "Swinging on pitchout",
        "pitch_counts": 1,
        "did_swing": True,
        "made_contact": False,
    },
    "R": {
        "description": "Foul ball on pitchout",
        "pitch_counts": 1,
        "did_swing": True,
        "made_contact": True,
    },
    "U": {
        "description": "Unknown or missed pitch",
        "pitch_counts": 1,
        "did_swing": False,
        "made_contact": False,
    },
    "V": {
        "description": "(Called ball because pitcher went to his mouth)",
        "pitch_counts": 0,
        "did_swing": False,
        "made_contact": False,
    },
    "Y": {
        "description": "Ball put into play on pitchout",
        "pitch_counts": 1,
        "did_swing": True,
        "made_contact": True,
    },
    "1": {
        "description": "(Pickoff throw to first)",
        "pitch_counts": 0,
        "did_swing": False,
        "made_contact": False,
    },
    "2": {
        "description": "(Pickoff throw to second)",
        "pitch_counts": 0,
        "did_swing": False,
        "made_contact": False,
    },
    "3": {
        "description": "(Pickoff throw to third)",
        "pitch_counts": 0,
        "did_swing": False,
        "made_contact": False,
    },
    ">": {
        "description": "(Runner going on the pitch)",
        "pitch_counts": 0,
        "did_swing": False,
        "made_contact": False,
    },
    "+": {
        "description": "(Pickoff throw by the catcher)",
        "pitch_counts": 0,
        "did_swing": False,
        "made_contact": False,
    },
    "*": {
        "description": "(The following pitch was blocked by the catcher)",
        "pitch_counts": 0,
        "did_swing": False,
        "made_contact": False,
    },
    ".": {
        "description": "(Play not involving the batter)",
        "pitch_counts": 0,
        "did_swing": False,
        "made_contact": False,
    },
}

VENUE_TERMS = ["stadium", "park", "field", "coliseum", "centre", "estadio", "dome"]

ZONE_LOCATIONS_INSIDE_SZ = [6, 7, 8, 11, 12, 13, 16, 17, 18]

PITCH_DES_SWINGING_STRIKE = [
    "Foul Tip",
    "Missed Bunt",
    "Swinging Pitchout",
    "Swinging Strike",
    "Swinging Strike (Blocked)",
]

PITCH_DES_DID_SWING = [
    "Foul",
    "Foul (Runner Going)",
    "Foul Bunt",
    "Foul Tip",
    "In play, no out",
    "In play; no out",
    "In play, out(s)",
    "In play; out(s)",
    "In play, run(s)",
    "In play; run(s)",
    "Missed Bunt",
    "Swinging Pitchout",
    "Swinging Strike",
    "Swinging Strike (Blocked)",
]

PITCH_DES_MADE_CONTACT = [
    "Foul",
    "Foul (Runner Going)",
    "Foul Bunt",
    "Foul Tip",
    "In play, no out",
    "In play; no out",
    "In play, out(s)",
    "In play; out(s)",
    "In play, run(s)",
    "In play; run(s)",
]

PLAY_DES_BB_TYPE_FB = ["Fly Ball", "Flyball", "Popfly"]
PLAY_DES_BB_TYPE_GB = ["Ground Ball", "Groundout"]
PLAY_DES_BB_TYPE_LD = ["Line Drive", "Lineout"]
PLAY_DES_BB_TYPE_PU = ["Popfly"]

AT_BAT_RESULTS_OUT = [
    "bunt groundout",
    "bunt pop out",
    "double play",
    "flyout",
    "forceout",
    "grounded into dp",
    "groundout",
    "lineout",
    "pop out",
    "runner out",
    "sac bunt",
    "sac fly",
    "sac fly dp",
    "sacrifice bunt dp",
    "strikeout",
    "strikeout - dp",
    "triple play",
]

AT_BAT_RESULTS_UNCLEAR = [
    "batter interference",
    "catcher interference",
    "fan interference",
    "fielders choice",
    "fielders choice out",
    "missing_des",
]

AT_BAT_RESULTS_HIT = [
    "double",
    "home run",
    "single",
    "triple",
]

AT_BAT_RESULTS_WALK = [
    "intent walk",
    "walk",
]

AT_BAT_RESULTS_STRIKEOUT = [
    "strikeout",
    "strikeout - dp",
]

AT_BAT_RESULTS_HBP = [
    "hit by pitch",
]

AT_BAT_RESULTS_ERROR = [
    "field error",
]

AT_BAT_RESULTS_SAC_HIT = [
    "sac bunt",
    "sacrifice bunt dp",
]

AT_BAT_RESULTS_SAC_FLY = [
    "sac fly",
    "sac fly dp",
]

JOB_SPINNER_COLORS = {
    DataSet.BBREF_GAMES_FOR_DATE: "red",
    DataSet.BROOKS_GAMES_FOR_DATE: "blue",
    DataSet.BBREF_BOXSCORES: "green",
    DataSet.BROOKS_PITCH_LOGS: "cyan",
    DataSet.BROOKS_PITCHFX: "magenta",
}

DATA_SET_FROM_NAME_MAP = {
    "Games for Date (bbref.com)": DataSet.BBREF_GAMES_FOR_DATE,
    "Games for Date (brooksbaseball.net)": DataSet.BROOKS_GAMES_FOR_DATE,
    "Boxscores (bbref.com)": DataSet.BBREF_BOXSCORES,
    "Pitch Logs for Game (brooksbaseball.net)": DataSet.BROOKS_PITCH_LOGS,
    "PitchFX Logs (brooksbaseball.net)": DataSet.BROOKS_PITCHFX,
}

DATA_SET_TO_NAME_MAP = {data_set: name for name, data_set in DATA_SET_FROM_NAME_MAP.items()}

CLI_COLORS = [
    "black",
    "red",
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan",
    "white",
    "bright_black",
    "bright_red",
    "bright_green",
    "bright_yellow",
    "bright_blue",
    "bright_magenta",
    "bright_cyan",
    "bright_white",
]

EMOJIS = {
    "HAND_POINTER": "👉",
    "BOOKMARK": "🔖",
    "FOLDER": "📁",
    "CHART": "📊",
    "BOOK": "📚",
    "TABBED_FILES": "🗂️",
    "TOOLS": "🛠️ ",
    "KNIFE": "🔪",
    "ROBOT": "🤖",
    "KNOBS": "🎛️",
    "BASKET": "🧺",
    "PACKAGE": "📦",
    "BAGS": "🛍️",
    "HONEY_POT": "🍯",
    "CALENDAR": "📅",
    "CLOCK": "🕚",
    "PAGER": "📟",
    "SHRUG": "🤷‍♂️",
    "CRYING": "😢",
    "WEARY": "😩",
    "UPSIDE_DOWN": "🙃",
    "CONFUSED": "😕",
    "THUMBS_UP": "👍",
    "CLOUD": "🌧",
    "SPIRAL": "🌀",
    "DIZZY": "💫",
    "BOLT": "⚡",
    "BOMB": "💣",
    "FLASHLIGHT": "🔦",
    "MICROSCOPE": "🔬",
    "NEWSPAPER": "📰",
    "CAP": "🧢",
    "BASEBALL": "⚾",
    "SOFTBALL": "🥎",
    "HOTDOG": "🌭",
    "CIRCLE": "🔵",
    "GEAR": "⚙️ ",
    "BANG": "💥",
    "BLUE_DIAMOND": "🔹",
    "CHECK": "☑️",
    "ASTERISK": "✳️",
    "UP": "⬆️",
    "DOWN": "⬇️",
    "LEFT": "⬅️",
    "RIGHT": "➡️",
    "PASSED": "✅",
    "FAILED": "❌",
    "QUESTION": "❔",
    "COOL": "🆒",
    "BACK": "👈",
    "EXIT": "❎",
}

MENU_NUMBERS = {
    1: "1️⃣",
    2: "2️⃣",
    3: "3️⃣",
    4: "4️⃣",
    5: "5️⃣",
    6: "6️⃣",
    7: "7️⃣",
    8: "8️⃣",
    9: "9️⃣",
    10: "🔟",
}

FAKE_SPINNER = {"interval": 140, "frames": ["⚡", "⚡"]}

FIGLET_FONTS = [
    "3-d",
    "acrobatic",
    "banner3-D",
    "basic",
    "charact6",
    "contrast",
    "cyberlarge",
    "doom",
    "double",
    "ebbs_1__",
    "eftiwater",
    "epic",
    "fender",
    "graffiti",
    "home_pak",
    "isometric1",
    "jazmine",
    "kban",
    "larry3d",
    "lean",
    "lockergnome",
    "nancyj-fancy",
    "os2",
    "peaks",
    "rowancap",
    "sblood",
    "shimrod",
    "slant",
    "smscript",
    "space_op",
    "speed",
    "stellar",
    "ticksslant",
    "tinker-toy",
    "tubular",
    "usaflag",
]



# flake8: noqa
from vigorish.views.bat_stats_view import (
    BatStats_All_View,
    BatStats_By_Opp_Team_View,
    BatStats_By_Opp_Team_Year_View,
    BatStats_By_Team_View,
    BatStats_By_Team_Year_View,
    BatStats_By_Year_View,
)
from vigorish.views.batter_pitchfx_view import (
    Batter_PitchFx_All_View,
    Batter_PitchFx_By_Year_View,
    Batter_PitchFx_For_Game_All_View,
    Batter_PitchFx_For_Game_vs_LHP_as_LHB_View,
    Batter_PitchFx_For_Game_vs_LHP_as_RHB_View,
    Batter_PitchFx_For_Game_vs_RHP_as_LHB_View,
    Batter_PitchFx_For_Game_vs_RHP_as_RHB_View,
    Batter_PitchFx_vs_LHP_as_LHB_Career_View,
    Batter_PitchFx_vs_LHP_as_RHB_Career_View,
    Batter_PitchFx_vs_RHP_as_LHB_Career_View,
    Batter_PitchFx_vs_RHP_as_RHB_Career_View,
    Batter_PitchType_All_View,
    Batter_PitchType_By_Year_View,
    Batter_PitchType_For_Game_All_View,
    Batter_PitchType_For_Game_vs_LHP_as_LHB_View,
    Batter_PitchType_For_Game_vs_LHP_as_RHB_View,
    Batter_PitchType_For_Game_vs_RHP_as_LHB_View,
    Batter_PitchType_For_Game_vs_RHP_as_RHB_View,
    Batter_PitchType_vs_LHP_as_LHB_Career_View,
    Batter_PitchType_vs_LHP_as_RHB_Career_View,
    Batter_PitchType_vs_RHP_as_LHB_Career_View,
    Batter_PitchType_vs_RHP_as_RHB_Career_View,
)
from vigorish.views.date_pitch_app_view import Date_PitchApp_View
from vigorish.views.game_pitch_app_view import Game_PitchApp_View
from vigorish.views.pitch_stats_view import (
    PitchStats_All_View,
    PitchStats_By_Opp_Team_By_Year_View,
    PitchStats_By_Opp_Team_View,
    PitchStats_By_Team_By_Year_View,
    PitchStats_By_Team_View,
    PitchStats_By_Year_View,
    PitchStats_RP_View,
    PitchStats_SP_View,
)
from vigorish.views.pitcher_pitchfx_view import (
    Pitcher_PitchFx_All_By_Year_View,
    Pitcher_PitchFx_All_View,
    Pitcher_PitchFx_For_Game_All_View,
    Pitcher_PitchFx_For_Game_vs_LHB_View,
    Pitcher_PitchFx_For_Game_vs_RHB_View,
    Pitcher_PitchFx_vs_LHB_By_Year_View,
    Pitcher_PitchFx_vs_LHB_View,
    Pitcher_PitchFx_vs_RHB_By_Year_View,
    Pitcher_PitchFx_vs_RHB_View,
    Pitcher_PitchType_All_By_Year_View,
    Pitcher_PitchType_All_View,
    Pitcher_PitchType_For_Game_All_View,
    Pitcher_PitchType_For_Game_vs_LHB_View,
    Pitcher_PitchType_For_Game_vs_RHB_View,
    Pitcher_PitchType_vs_LHB_By_Year_View,
    Pitcher_PitchType_vs_LHB_View,
    Pitcher_PitchType_vs_RHB_By_Year_View,
    Pitcher_PitchType_vs_RHB_View,
)
from vigorish.views.pitchtype_pitchfx_view import (
    PitchType_All_View,
    PitchType_Left_View,
    PitchType_Right_View,
)
from vigorish.views.season_date_view import Season_Date_View
from vigorish.views.season_game_pitch_app_view import Season_Game_PitchApp_View
from vigorish.views.season_game_view import Season_Game_View
from vigorish.views.season_pitch_app_view import Season_PitchApp_View
from vigorish.views.team_bat_stats_view import (
    Team_BatStats_By_BatOrder_By_Player_By_Year,
    Team_BatStats_By_BatOrder_By_Year,
    Team_BatStats_By_DefPosition_By_Player_By_Year,
    Team_BatStats_By_DefPosition_By_Year,
    Team_BatStats_By_Player_By_Year_View,
    Team_BatStats_By_Year_View,
    Team_BatStats_For_Starters_By_Player_By_Year,
    Team_BatStats_For_Starters_By_Year,
    Team_BatStats_For_Subs_By_Player_By_Year,
    Team_BatStats_For_Subs_By_Year,
)
from vigorish.views.team_pitch_stats_view import (
    Team_PitchStats_By_Player_By_Year_View,
    Team_PitchStats_By_Year_View,
    Team_PitchStats_RP_By_Player_By_Year_View,
    Team_PitchStats_RP_By_Year_View,
    Team_PitchStats_SP_By_Player_By_Year_View,
    Team_PitchStats_SP_By_Year_View,
)



from sqlalchemy import and_, func, join, literal, or_, select
from sqlalchemy_utils import create_view

import vigorish.database as db
from vigorish.views.pfx_col_expressions import (
    avg,
    bb_rate,
    called_strike_rate,
    contact_rate,
    csw_rate,
    fly_ball_rate,
    ground_ball_rate,
    hr_per_fb,
    iso,
    k_rate,
    line_drive_rate,
    o_contact_rate,
    o_swing_rate,
    obp,
    ops,
    pop_up_rate,
    slg,
    swing_rate,
    swinging_strike_rate,
    total_at_bats,
    whiff_rate,
    z_contact_rate,
    z_swing_rate,
    zone_rate,
)


class Batter_PitchFx_All_View(db.Base):
    __table__ = create_view(
        name="batter_pitchfx_all",
        selectable=select(
            [
                db.PitchFx.batter_id.label("id"),
                db.PitchFx.batter_id_mlb.label("mlb_id"),
                literal("ALL").label("pitch_type"),
                func.count(db.PitchFx.id).label("total_pitches"),
                func.sum(db.PitchFx.is_final_pitch_of_ab).label("total_pa"),
                total_at_bats,
                func.sum(db.PitchFx.ab_result_out).label("total_outs"),
                func.sum(db.PitchFx.ab_result_hit).label("total_hits"),
                func.sum(db.PitchFx.ab_result_bb).label("total_bb"),
                func.sum(db.PitchFx.ab_result_k).label("total_k"),
                avg,
                obp,
                slg,
                ops,
                iso,
                fly_ball_rate,
                ground_ball_rate,
                line_drive_rate,
                pop_up_rate,
                bb_rate,
                k_rate,
                hr_per_fb,
                zone_rate,
                called_strike_rate,
                swinging_strike_rate,
                whiff_rate,
                csw_rate,
                o_swing_rate,
                z_swing_rate,
                swing_rate,
                o_contact_rate,
                z_contact_rate,
                contact_rate,
                func.sum(db.PitchFx.batter_did_swing).label("total_swings"),
                func.sum(db.PitchFx.batter_made_contact).label("total_swings_made_contact"),
                func.sum(db.PitchFx.called_strike).label("total_called_strikes"),
                func.sum(db.PitchFx.swinging_strike).label("total_swinging_strikes"),
                func.sum(db.PitchFx.inside_strike_zone).label("total_inside_strike_zone"),
                func.sum(db.PitchFx.outside_strike_zone).label("total_outside_strike_zone"),
                func.sum(db.PitchFx.swing_inside_zone).label("total_swings_inside_zone"),
                func.sum(db.PitchFx.swing_outside_zone).label("total_swings_outside_zone"),
                func.sum(db.PitchFx.contact_inside_zone).label("total_contact_inside_zone"),
                func.sum(db.PitchFx.contact_outside_zone).label("total_contact_outside_zone"),
                func.sum(db.PitchFx.is_batted_ball).label("total_batted_balls"),
                func.sum(db.PitchFx.is_ground_ball).label("total_ground_balls"),
                func.sum(db.PitchFx.is_line_drive).label("total_line_drives"),
                func.sum(db.PitchFx.is_fly_ball).label("total_fly_balls"),
                func.sum(db.PitchFx.is_pop_up).label("total_pop_ups"),
                func.sum(db.PitchFx.ab_result_single).label("total_singles"),
                func.sum(db.PitchFx.ab_result_double).label("total_doubles"),
                func.sum(db.PitchFx.ab_result_triple).label("total_triples"),
                func.sum(db.PitchFx.ab_result_homerun).label("total_homeruns"),
                func.sum(db.PitchFx.ab_result_ibb).label("total_ibb"),
                func.sum(db.PitchFx.ab_result_hbp).label("total_hbp"),
                func.sum(db.PitchFx.ab_result_error).label("total_errors"),
                func.sum(db.PitchFx.ab_result_sac_hit).label("total_sac_hit"),
                func.sum(db.PitchFx.ab_result_sac_fly).label("total_sac_fly"),
                func.sum(db.PitchFx.ab_result_unclear).label("total_unclear_results"),
                func.sum(db.PitchFx.pitch_type_int.distinct()).label("pitch_type_int"),
            ]
        )
        .where(
            and_(
                db.PitchFx.is_invalid_ibb == 0,
                db.PitchFx.is_out_of_sequence == 0,
                or_(db.PitchFx.stand == "L", db.PitchFx.stand == "R"),
                or_(db.PitchFx.p_throws == "L", db.PitchFx.p_throws == "R"),
            )
        )
        .select_from(db.PitchFx)
        .group_by(db.PitchFx.batter_id)
        .order_by(db.PitchFx.batter_id_mlb),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pfx_metrics_for_career_for_batter(cls, db_engine, mlb_id):
        results = db_engine.execute(select([cls]).where(cls.mlb_id == mlb_id)).fetchall()
        return [dict(row) for row in results][0] if results else {}


class Batter_PitchType_All_View(db.Base):
    __table__ = create_view(
        name="batter_pitchtype_all",
        selectable=select(
            [
                db.PitchFx.batter_id.label("id"),
                db.PitchFx.batter_id_mlb.label("mlb_id"),
                db.PitchFx.mlbam_pitch_name.label("pitch_type"),
                func.count(db.PitchFx.id).label("total_pitches"),
                func.sum(db.PitchFx.is_final_pitch_of_ab).label("total_pa"),
                total_at_bats,
                func.sum(db.PitchFx.ab_result_out).label("total_outs"),
                func.sum(db.PitchFx.ab_result_hit).label("total_hits"),
                func.sum(db.PitchFx.ab_result_bb).label("total_bb"),
                func.sum(db.PitchFx.ab_result_k).label("total_k"),
                func.avg(db.PitchFx.start_speed).label("avg_speed"),
                func.avg(db.PitchFx.pfx_x).label("avg_pfx_x"),
                func.avg(db.PitchFx.pfx_z).label("avg_pfx_z"),
                func.avg(db.PitchFx.px).label("avg_px"),
                func.avg(db.PitchFx.pz).label("avg_pz"),
                avg,
                obp,
                slg,
                ops,
                iso,
                fly_ball_rate,
                ground_ball_rate,
                line_drive_rate,
                pop_up_rate,
                bb_rate,
                k_rate,
                hr_per_fb,
                zone_rate,
                called_strike_rate,
                swinging_strike_rate,
                whiff_rate,
                csw_rate,
                o_swing_rate,
                z_swing_rate,
                swing_rate,
                o_contact_rate,
                z_contact_rate,
                contact_rate,
                func.sum(db.PitchFx.batter_did_swing).label("total_swings"),
                func.sum(db.PitchFx.batter_made_contact).label("total_swings_made_contact"),
                func.sum(db.PitchFx.called_strike).label("total_called_strikes"),
                func.sum(db.PitchFx.swinging_strike).label("total_swinging_strikes"),
                func.sum(db.PitchFx.inside_strike_zone).label("total_inside_strike_zone"),
                func.sum(db.PitchFx.outside_strike_zone).label("total_outside_strike_zone"),
                func.sum(db.PitchFx.swing_inside_zone).label("total_swings_inside_zone"),
                func.sum(db.PitchFx.swing_outside_zone).label("total_swings_outside_zone"),
                func.sum(db.PitchFx.contact_inside_zone).label("total_contact_inside_zone"),
                func.sum(db.PitchFx.contact_outside_zone).label("total_contact_outside_zone"),
                func.sum(db.PitchFx.is_batted_ball).label("total_batted_balls"),
                func.sum(db.PitchFx.is_ground_ball).label("total_ground_balls"),
                func.sum(db.PitchFx.is_line_drive).label("total_line_drives"),
                func.sum(db.PitchFx.is_fly_ball).label("total_fly_balls"),
                func.sum(db.PitchFx.is_pop_up).label("total_pop_ups"),
                func.sum(db.PitchFx.ab_result_single).label("total_singles"),
                func.sum(db.PitchFx.ab_result_double).label("total_doubles"),
                func.sum(db.PitchFx.ab_result_triple).label("total_triples"),
                func.sum(db.PitchFx.ab_result_homerun).label("total_homeruns"),
                func.sum(db.PitchFx.ab_result_ibb).label("total_ibb"),
                func.sum(db.PitchFx.ab_result_hbp).label("total_hbp"),
                func.sum(db.PitchFx.ab_result_error).label("total_errors"),
                func.sum(db.PitchFx.ab_result_sac_hit).label("total_sac_hit"),
                func.sum(db.PitchFx.ab_result_sac_fly).label("total_sac_fly"),
                func.sum(db.PitchFx.ab_result_unclear).label("total_unclear_results"),
            ]
        )
        .where(
            and_(
                db.PitchFx.is_invalid_ibb == 0,
                db.PitchFx.is_out_of_sequence == 0,
                or_(db.PitchFx.stand == "L", db.PitchFx.stand == "R"),
                or_(db.PitchFx.p_throws == "L", db.PitchFx.p_throws == "R"),
            )
        )
        .select_from(db.PitchFx)
        .group_by(db.PitchFx.batter_id)
        .group_by(db.PitchFx.mlbam_pitch_name)
        .order_by(db.PitchFx.batter_id_mlb),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pfx_metrics_for_career_for_batter(cls, db_engine, mlb_id):
        results = db_engine.execute(select([cls]).where(cls.mlb_id == mlb_id)).fetchall()
        return [dict(row) for row in results] if results else []


class Batter_PitchFx_vs_RHP_as_RHB_Career_View(db.Base):
    __table__ = create_view(
        name="batter_pitchfx_vs_rhp_as_rhb_career",
        selectable=select(
            [
                db.PitchFx.batter_id.label("id"),
                db.PitchFx.batter_id_mlb.label("mlb_id"),
                db.PitchFx.stand.label("bat_stand"),
                db.PitchFx.p_throws.label("p_throws"),
                literal("ALL").label("pitch_type"),
                func.count(db.PitchFx.id).label("total_pitches"),
                func.sum(db.PitchFx.is_final_pitch_of_ab).label("total_pa"),
                total_at_bats,
                func.sum(db.PitchFx.ab_result_out).label("total_outs"),
                func.sum(db.PitchFx.ab_result_hit).label("total_hits"),
                func.sum(db.PitchFx.ab_result_bb).label("total_bb"),
                func.sum(db.PitchFx.ab_result_k).label("total_k"),
                avg,
                obp,
                slg,
                ops,
                iso,
                fly_ball_rate,
                ground_ball_rate,
                line_drive_rate,
                pop_up_rate,
                bb_rate,
                k_rate,
                hr_per_fb,
                zone_rate,
                called_strike_rate,
                swinging_strike_rate,
                whiff_rate,
                csw_rate,
                o_swing_rate,
                z_swing_rate,
                swing_rate,
                o_contact_rate,
                z_contact_rate,
                contact_rate,
                func.sum(db.PitchFx.batter_did_swing).label("total_swings"),
                func.sum(db.PitchFx.batter_made_contact).label("total_swings_made_contact"),
                func.sum(db.PitchFx.called_strike).label("total_called_strikes"),
                func.sum(db.PitchFx.swinging_strike).label("total_swinging_strikes"),
                func.sum(db.PitchFx.inside_strike_zone).label("total_inside_strike_zone"),
                func.sum(db.PitchFx.outside_strike_zone).label("total_outside_strike_zone"),
                func.sum(db.PitchFx.swing_inside_zone).label("total_swings_inside_zone"),
                func.sum(db.PitchFx.swing_outside_zone).label("total_swings_outside_zone"),
                func.sum(db.PitchFx.contact_inside_zone).label("total_contact_inside_zone"),
                func.sum(db.PitchFx.contact_outside_zone).label("total_contact_outside_zone"),
                func.sum(db.PitchFx.is_batted_ball).label("total_batted_balls"),
                func.sum(db.PitchFx.is_ground_ball).label("total_ground_balls"),
                func.sum(db.PitchFx.is_line_drive).label("total_line_drives"),
                func.sum(db.PitchFx.is_fly_ball).label("total_fly_balls"),
                func.sum(db.PitchFx.is_pop_up).label("total_pop_ups"),
                func.sum(db.PitchFx.ab_result_single).label("total_singles"),
                func.sum(db.PitchFx.ab_result_double).label("total_doubles"),
                func.sum(db.PitchFx.ab_result_triple).label("total_triples"),
                func.sum(db.PitchFx.ab_result_homerun).label("total_homeruns"),
                func.sum(db.PitchFx.ab_result_ibb).label("total_ibb"),
                func.sum(db.PitchFx.ab_result_hbp).label("total_hbp"),
                func.sum(db.PitchFx.ab_result_error).label("total_errors"),
                func.sum(db.PitchFx.ab_result_sac_hit).label("total_sac_hit"),
                func.sum(db.PitchFx.ab_result_sac_fly).label("total_sac_fly"),
                func.sum(db.PitchFx.ab_result_unclear).label("total_unclear_results"),
                func.sum(db.PitchFx.pitch_type_int.distinct()).label("pitch_type_int"),
            ]
        )
        .where(
            and_(
                db.PitchFx.is_invalid_ibb == 0,
                db.PitchFx.is_out_of_sequence == 0,
                db.PitchFx.stand == "R",
                db.PitchFx.p_throws == "R",
            )
        )
        .select_from(db.PitchFx)
        .group_by(db.PitchFx.batter_id)
        .order_by(db.PitchFx.batter_id_mlb),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pfx_metrics_vs_rhp_as_rhb_for_career_for_batter(cls, db_engine, mlb_id):
        results = db_engine.execute(select([cls]).where(cls.mlb_id == mlb_id)).fetchall()
        return [dict(row) for row in results][0] if results else {}


class Batter_PitchType_vs_RHP_as_RHB_Career_View(db.Base):
    __table__ = create_view(
        name="batter_pitchtype_vs_rhp_as_rhb_career",
        selectable=select(
            [
                db.PitchFx.batter_id.label("id"),
                db.PitchFx.batter_id_mlb.label("mlb_id"),
                db.PitchFx.stand.label("bat_stand"),
                db.PitchFx.p_throws.label("p_throws"),
                db.PitchFx.mlbam_pitch_name.label("pitch_type"),
                func.count(db.PitchFx.id).label("total_pitches"),
                func.sum(db.PitchFx.is_final_pitch_of_ab).label("total_pa"),
                total_at_bats,
                func.sum(db.PitchFx.ab_result_out).label("total_outs"),
                func.sum(db.PitchFx.ab_result_hit).label("total_hits"),
                func.sum(db.PitchFx.ab_result_bb).label("total_bb"),
                func.sum(db.PitchFx.ab_result_k).label("total_k"),
                avg,
                obp,
                slg,
                ops,
                iso,
                fly_ball_rate,
                ground_ball_rate,
                line_drive_rate,
                pop_up_rate,
                bb_rate,
                k_rate,
                hr_per_fb,
                zone_rate,
                called_strike_rate,
                swinging_strike_rate,
                whiff_rate,
                csw_rate,
                o_swing_rate,
                z_swing_rate,
                swing_rate,
                o_contact_rate,
                z_contact_rate,
                contact_rate,
                func.sum(db.PitchFx.batter_did_swing).label("total_swings"),
                func.sum(db.PitchFx.batter_made_contact).label("total_swings_made_contact"),
                func.sum(db.PitchFx.called_strike).label("total_called_strikes"),
                func.sum(db.PitchFx.swinging_strike).label("total_swinging_strikes"),
                func.sum(db.PitchFx.inside_strike_zone).label("total_inside_strike_zone"),
                func.sum(db.PitchFx.outside_strike_zone).label("total_outside_strike_zone"),
                func.sum(db.PitchFx.swing_inside_zone).label("total_swings_inside_zone"),
                func.sum(db.PitchFx.swing_outside_zone).label("total_swings_outside_zone"),
                func.sum(db.PitchFx.contact_inside_zone).label("total_contact_inside_zone"),
                func.sum(db.PitchFx.contact_outside_zone).label("total_contact_outside_zone"),
                func.sum(db.PitchFx.is_batted_ball).label("total_batted_balls"),
                func.sum(db.PitchFx.is_ground_ball).label("total_ground_balls"),
                func.sum(db.PitchFx.is_line_drive).label("total_line_drives"),
                func.sum(db.PitchFx.is_fly_ball).label("total_fly_balls"),
                func.sum(db.PitchFx.is_pop_up).label("total_pop_ups"),
                func.sum(db.PitchFx.ab_result_single).label("total_singles"),
                func.sum(db.PitchFx.ab_result_double).label("total_doubles"),
                func.sum(db.PitchFx.ab_result_triple).label("total_triples"),
                func.sum(db.PitchFx.ab_result_homerun).label("total_homeruns"),
                func.sum(db.PitchFx.ab_result_ibb).label("total_ibb"),
                func.sum(db.PitchFx.ab_result_hbp).label("total_hbp"),
                func.sum(db.PitchFx.ab_result_error).label("total_errors"),
                func.sum(db.PitchFx.ab_result_sac_hit).label("total_sac_hit"),
                func.sum(db.PitchFx.ab_result_sac_fly).label("total_sac_fly"),
                func.sum(db.PitchFx.ab_result_unclear).label("total_unclear_results"),
                func.sum(db.PitchFx.pitch_type_int.distinct()).label("pitch_type_int"),
            ]
        )
        .where(
            and_(
                db.PitchFx.is_invalid_ibb == 0,
                db.PitchFx.is_out_of_sequence == 0,
                db.PitchFx.stand == "R",
                db.PitchFx.p_throws == "R",
            )
        )
        .select_from(db.PitchFx)
        .group_by(db.PitchFx.batter_id)
        .group_by(db.PitchFx.mlbam_pitch_name)
        .order_by(db.PitchFx.batter_id_mlb),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pfx_metrics_vs_rhp_as_rhb_for_career_for_batter(cls, db_engine, mlb_id):
        results = db_engine.execute(select([cls]).where(cls.mlb_id == mlb_id)).fetchall()
        return [dict(row) for row in results] if results else []


class Batter_PitchFx_vs_RHP_as_LHB_Career_View(db.Base):
    __table__ = create_view(
        name="batter_pitchfx_vs_rhp_as_lhb_career",
        selectable=select(
            [
                db.PitchFx.batter_id.label("id"),
                db.PitchFx.batter_id_mlb.label("mlb_id"),
                db.PitchFx.stand.label("bat_stand"),
                db.PitchFx.p_throws.label("p_throws"),
                literal("ALL").label("pitch_type"),
                func.count(db.PitchFx.id).label("total_pitches"),
                func.sum(db.PitchFx.is_final_pitch_of_ab).label("total_pa"),
                total_at_bats,
                func.sum(db.PitchFx.ab_result_out).label("total_outs"),
                func.sum(db.PitchFx.ab_result_hit).label("total_hits"),
                func.sum(db.PitchFx.ab_result_bb).label("total_bb"),
                func.sum(db.PitchFx.ab_result_k).label("total_k"),
                avg,
                obp,
                slg,
                ops,
                iso,
                fly_ball_rate,
                ground_ball_rate,
                line_drive_rate,
                pop_up_rate,
                bb_rate,
                k_rate,
                hr_per_fb,
                zone_rate,
                called_strike_rate,
                swinging_strike_rate,
                whiff_rate,
                csw_rate,
                o_swing_rate,
                z_swing_rate,
                swing_rate,
                o_contact_rate,
                z_contact_rate,
                contact_rate,
                func.sum(db.PitchFx.batter_did_swing).label("total_swings"),
                func.sum(db.PitchFx.batter_made_contact).label("total_swings_made_contact"),
                func.sum(db.PitchFx.called_strike).label("total_called_strikes"),
                func.sum(db.PitchFx.swinging_strike).label("total_swinging_strikes"),
                func.sum(db.PitchFx.inside_strike_zone).label("total_inside_strike_zone"),
                func.sum(db.PitchFx.outside_strike_zone).label("total_outside_strike_zone"),
                func.sum(db.PitchFx.swing_inside_zone).label("total_swings_inside_zone"),
                func.sum(db.PitchFx.swing_outside_zone).label("total_swings_outside_zone"),
                func.sum(db.PitchFx.contact_inside_zone).label("total_contact_inside_zone"),
                func.sum(db.PitchFx.contact_outside_zone).label("total_contact_outside_zone"),
                func.sum(db.PitchFx.is_batted_ball).label("total_batted_balls"),
                func.sum(db.PitchFx.is_ground_ball).label("total_ground_balls"),
                func.sum(db.PitchFx.is_line_drive).label("total_line_drives"),
                func.sum(db.PitchFx.is_fly_ball).label("total_fly_balls"),
                func.sum(db.PitchFx.is_pop_up).label("total_pop_ups"),
                func.sum(db.PitchFx.ab_result_single).label("total_singles"),
                func.sum(db.PitchFx.ab_result_double).label("total_doubles"),
                func.sum(db.PitchFx.ab_result_triple).label("total_triples"),
                func.sum(db.PitchFx.ab_result_homerun).label("total_homeruns"),
                func.sum(db.PitchFx.ab_result_ibb).label("total_ibb"),
                func.sum(db.PitchFx.ab_result_hbp).label("total_hbp"),
                func.sum(db.PitchFx.ab_result_error).label("total_errors"),
                func.sum(db.PitchFx.ab_result_sac_hit).label("total_sac_hit"),
                func.sum(db.PitchFx.ab_result_sac_fly).label("total_sac_fly"),
                func.sum(db.PitchFx.ab_result_unclear).label("total_unclear_results"),
                func.sum(db.PitchFx.pitch_type_int.distinct()).label("pitch_type_int"),
            ]
        )
        .where(
            and_(
                db.PitchFx.is_invalid_ibb == 0,
                db.PitchFx.is_out_of_sequence == 0,
                db.PitchFx.stand == "L",
                db.PitchFx.p_throws == "R",
            )
        )
        .select_from(db.PitchFx)
        .group_by(db.PitchFx.batter_id)
        .order_by(db.PitchFx.batter_id_mlb),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pfx_metrics_vs_rhp_as_lhb_for_career_for_batter(cls, db_engine, mlb_id):
        results = db_engine.execute(select([cls]).where(cls.mlb_id == mlb_id)).fetchall()
        return [dict(row) for row in results][0] if results else {}


class Batter_PitchType_vs_RHP_as_LHB_Career_View(db.Base):
    __table__ = create_view(
        name="batter_pitchtype_vs_rhp_as_lhb_career",
        selectable=select(
            [
                db.PitchFx.batter_id.label("id"),
                db.PitchFx.batter_id_mlb.label("mlb_id"),
                db.PitchFx.stand.label("bat_stand"),
                db.PitchFx.p_throws.label("p_throws"),
                db.PitchFx.mlbam_pitch_name.label("pitch_type"),
                func.count(db.PitchFx.id).label("total_pitches"),
                func.sum(db.PitchFx.is_final_pitch_of_ab).label("total_pa"),
                total_at_bats,
                func.sum(db.PitchFx.ab_result_out).label("total_outs"),
                func.sum(db.PitchFx.ab_result_hit).label("total_hits"),
                func.sum(db.PitchFx.ab_result_bb).label("total_bb"),
                func.sum(db.PitchFx.ab_result_k).label("total_k"),
                avg,
                obp,
                slg,
                ops,
                iso,
                fly_ball_rate,
                ground_ball_rate,
                line_drive_rate,
                pop_up_rate,
                bb_rate,
                k_rate,
                hr_per_fb,
                zone_rate,
                called_strike_rate,
                swinging_strike_rate,
                whiff_rate,
                csw_rate,
                o_swing_rate,
                z_swing_rate,
                swing_rate,
                o_contact_rate,
                z_contact_rate,
                contact_rate,
                func.sum(db.PitchFx.batter_did_swing).label("total_swings"),
                func.sum(db.PitchFx.batter_made_contact).label("total_swings_made_contact"),
                func.sum(db.PitchFx.called_strike).label("total_called_strikes"),
                func.sum(db.PitchFx.swinging_strike).label("total_swinging_strikes"),
                func.sum(db.PitchFx.inside_strike_zone).label("total_inside_strike_zone"),
                func.sum(db.PitchFx.outside_strike_zone).label("total_outside_strike_zone"),
                func.sum(db.PitchFx.swing_inside_zone).label("total_swings_inside_zone"),
                func.sum(db.PitchFx.swing_outside_zone).label("total_swings_outside_zone"),
                func.sum(db.PitchFx.contact_inside_zone).label("total_contact_inside_zone"),
                func.sum(db.PitchFx.contact_outside_zone).label("total_contact_outside_zone"),
                func.sum(db.PitchFx.is_batted_ball).label("total_batted_balls"),
                func.sum(db.PitchFx.is_ground_ball).label("total_ground_balls"),
                func.sum(db.PitchFx.is_line_drive).label("total_line_drives"),
                func.sum(db.PitchFx.is_fly_ball).label("total_fly_balls"),
                func.sum(db.PitchFx.is_pop_up).label("total_pop_ups"),
                func.sum(db.PitchFx.ab_result_single).label("total_singles"),
                func.sum(db.PitchFx.ab_result_double).label("total_doubles"),
                func.sum(db.PitchFx.ab_result_triple).label("total_triples"),
                func.sum(db.PitchFx.ab_result_homerun).label("total_homeruns"),
                func.sum(db.PitchFx.ab_result_ibb).label("total_ibb"),
                func.sum(db.PitchFx.ab_result_hbp).label("total_hbp"),
                func.sum(db.PitchFx.ab_result_error).label("total_errors"),
                func.sum(db.PitchFx.ab_result_sac_hit).label("total_sac_hit"),
                func.sum(db.PitchFx.ab_result_sac_fly).label("total_sac_fly"),
                func.sum(db.PitchFx.ab_result_unclear).label("total_unclear_results"),
                func.sum(db.PitchFx.pitch_type_int.distinct()).label("pitch_type_int"),
            ]
        )
        .where(
            and_(
                db.PitchFx.is_invalid_ibb == 0,
                db.PitchFx.is_out_of_sequence == 0,
                db.PitchFx.stand == "L",
                db.PitchFx.p_throws == "R",
            )
        )
        .select_from(db.PitchFx)
        .group_by(db.PitchFx.batter_id)
        .group_by(db.PitchFx.mlbam_pitch_name)
        .order_by(db.PitchFx.batter_id_mlb),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pfx_metrics_vs_rhp_as_lhb_for_career_for_batter(cls, db_engine, mlb_id):
        results = db_engine.execute(select([cls]).where(cls.mlb_id == mlb_id)).fetchall()
        return [dict(row) for row in results] if results else []


class Batter_PitchFx_vs_LHP_as_LHB_Career_View(db.Base):
    __table__ = create_view(
        name="batter_pitchfx_vs_lhp_as_lhb_career",
        selectable=select(
            [
                db.PitchFx.batter_id.label("id"),
                db.PitchFx.batter_id_mlb.label("mlb_id"),
                db.PitchFx.stand.label("bat_stand"),
                db.PitchFx.p_throws.label("p_throws"),
                literal("ALL").label("pitch_type"),
                func.count(db.PitchFx.id).label("total_pitches"),
                func.sum(db.PitchFx.is_final_pitch_of_ab).label("total_pa"),
                total_at_bats,
                func.sum(db.PitchFx.ab_result_out).label("total_outs"),
                func.sum(db.PitchFx.ab_result_hit).label("total_hits"),
                func.sum(db.PitchFx.ab_result_bb).label("total_bb"),
                func.sum(db.PitchFx.ab_result_k).label("total_k"),
                avg,
                obp,
                slg,
                ops,
                iso,
                fly_ball_rate,
                ground_ball_rate,
                line_drive_rate,
                pop_up_rate,
                bb_rate,
                k_rate,
                hr_per_fb,
                zone_rate,
                called_strike_rate,
                swinging_strike_rate,
                whiff_rate,
                csw_rate,
                o_swing_rate,
                z_swing_rate,
                swing_rate,
                o_contact_rate,
                z_contact_rate,
                contact_rate,
                func.sum(db.PitchFx.batter_did_swing).label("total_swings"),
                func.sum(db.PitchFx.batter_made_contact).label("total_swings_made_contact"),
                func.sum(db.PitchFx.called_strike).label("total_called_strikes"),
                func.sum(db.PitchFx.swinging_strike).label("total_swinging_strikes"),
                func.sum(db.PitchFx.inside_strike_zone).label("total_inside_strike_zone"),
                func.sum(db.PitchFx.outside_strike_zone).label("total_outside_strike_zone"),
                func.sum(db.PitchFx.swing_inside_zone).label("total_swings_inside_zone"),
                func.sum(db.PitchFx.swing_outside_zone).label("total_swings_outside_zone"),
                func.sum(db.PitchFx.contact_inside_zone).label("total_contact_inside_zone"),
                func.sum(db.PitchFx.contact_outside_zone).label("total_contact_outside_zone"),
                func.sum(db.PitchFx.is_batted_ball).label("total_batted_balls"),
                func.sum(db.PitchFx.is_ground_ball).label("total_ground_balls"),
                func.sum(db.PitchFx.is_line_drive).label("total_line_drives"),
                func.sum(db.PitchFx.is_fly_ball).label("total_fly_balls"),
                func.sum(db.PitchFx.is_pop_up).label("total_pop_ups"),
                func.sum(db.PitchFx.ab_result_single).label("total_singles"),
                func.sum(db.PitchFx.ab_result_double).label("total_doubles"),
                func.sum(db.PitchFx.ab_result_triple).label("total_triples"),
                func.sum(db.PitchFx.ab_result_homerun).label("total_homeruns"),
                func.sum(db.PitchFx.ab_result_ibb).label("total_ibb"),
                func.sum(db.PitchFx.ab_result_hbp).label("total_hbp"),
                func.sum(db.PitchFx.ab_result_error).label("total_errors"),
                func.sum(db.PitchFx.ab_result_sac_hit).label("total_sac_hit"),
                func.sum(db.PitchFx.ab_result_sac_fly).label("total_sac_fly"),
                func.sum(db.PitchFx.ab_result_unclear).label("total_unclear_results"),
                func.sum(db.PitchFx.pitch_type_int.distinct()).label("pitch_type_int"),
            ]
        )
        .where(
            and_(
                db.PitchFx.is_invalid_ibb == 0,
                db.PitchFx.is_out_of_sequence == 0,
                db.PitchFx.stand == "L",
                db.PitchFx.p_throws == "L",
            )
        )
        .select_from(db.PitchFx)
        .group_by(db.PitchFx.batter_id)
        .order_by(db.PitchFx.batter_id_mlb),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pfx_metrics_vs_lhp_as_lhb_for_career_for_batter(cls, db_engine, mlb_id):
        results = db_engine.execute(select([cls]).where(cls.mlb_id == mlb_id)).fetchall()
        return [dict(row) for row in results][0] if results else {}


class Batter_PitchType_vs_LHP_as_LHB_Career_View(db.Base):
    __table__ = create_view(
        name="batter_pitchtype_vs_lhp_as_lhb_career",
        selectable=select(
            [
                db.PitchFx.batter_id.label("id"),
                db.PitchFx.batter_id_mlb.label("mlb_id"),
                db.PitchFx.stand.label("bat_stand"),
                db.PitchFx.p_throws.label("p_throws"),
                db.PitchFx.mlbam_pitch_name.label("pitch_type"),
                func.count(db.PitchFx.id).label("total_pitches"),
                func.sum(db.PitchFx.is_final_pitch_of_ab).label("total_pa"),
                total_at_bats,
                func.sum(db.PitchFx.ab_result_out).label("total_outs"),
                func.sum(db.PitchFx.ab_result_hit).label("total_hits"),
                func.sum(db.PitchFx.ab_result_bb).label("total_bb"),
                func.sum(db.PitchFx.ab_result_k).label("total_k"),
                avg,
                obp,
                slg,
                ops,
                iso,
                fly_ball_rate,
                ground_ball_rate,
                line_drive_rate,
                pop_up_rate,
                bb_rate,
                k_rate,
                hr_per_fb,
                zone_rate,
                called_strike_rate,
                swinging_strike_rate,
                whiff_rate,
                csw_rate,
                o_swing_rate,
                z_swing_rate,
                swing_rate,
                o_contact_rate,
                z_contact_rate,
                contact_rate,
                func.sum(db.PitchFx.batter_did_swing).label("total_swings"),
                func.sum(db.PitchFx.batter_made_contact).label("total_swings_made_contact"),
                func.sum(db.PitchFx.called_strike).label("total_called_strikes"),
                func.sum(db.PitchFx.swinging_strike).label("total_swinging_strikes"),
                func.sum(db.PitchFx.inside_strike_zone).label("total_inside_strike_zone"),
                func.sum(db.PitchFx.outside_strike_zone).label("total_outside_strike_zone"),
                func.sum(db.PitchFx.swing_inside_zone).label("total_swings_inside_zone"),
                func.sum(db.PitchFx.swing_outside_zone).label("total_swings_outside_zone"),
                func.sum(db.PitchFx.contact_inside_zone).label("total_contact_inside_zone"),
                func.sum(db.PitchFx.contact_outside_zone).label("total_contact_outside_zone"),
                func.sum(db.PitchFx.is_batted_ball).label("total_batted_balls"),
                func.sum(db.PitchFx.is_ground_ball).label("total_ground_balls"),
                func.sum(db.PitchFx.is_line_drive).label("total_line_drives"),
                func.sum(db.PitchFx.is_fly_ball).label("total_fly_balls"),
                func.sum(db.PitchFx.is_pop_up).label("total_pop_ups"),
                func.sum(db.PitchFx.ab_result_single).label("total_singles"),
                func.sum(db.PitchFx.ab_result_double).label("total_doubles"),
                func.sum(db.PitchFx.ab_result_triple).label("total_triples"),
                func.sum(db.PitchFx.ab_result_homerun).label("total_homeruns"),
                func.sum(db.PitchFx.ab_result_ibb).label("total_ibb"),
                func.sum(db.PitchFx.ab_result_hbp).label("total_hbp"),
                func.sum(db.PitchFx.ab_result_error).label("total_errors"),
                func.sum(db.PitchFx.ab_result_sac_hit).label("total_sac_hit"),
                func.sum(db.PitchFx.ab_result_sac_fly).label("total_sac_fly"),
                func.sum(db.PitchFx.ab_result_unclear).label("total_unclear_results"),
                func.sum(db.PitchFx.pitch_type_int.distinct()).label("pitch_type_int"),
            ]
        )
        .where(
            and_(
                db.PitchFx.is_invalid_ibb == 0,
                db.PitchFx.is_out_of_sequence == 0,
                db.PitchFx.stand == "L",
                db.PitchFx.p_throws == "L",
            )
        )
        .select_from(db.PitchFx)
        .group_by(db.PitchFx.batter_id)
        .group_by(db.PitchFx.mlbam_pitch_name)
        .order_by(db.PitchFx.batter_id_mlb),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pfx_metrics_vs_lhp_as_lhb_for_career_for_batter(cls, db_engine, mlb_id):
        results = db_engine.execute(select([cls]).where(cls.mlb_id == mlb_id)).fetchall()
        return [dict(row) for row in results] if results else []


class Batter_PitchFx_vs_LHP_as_RHB_Career_View(db.Base):
    __table__ = create_view(
        name="batter_pitchfx_vs_lhp_as_rhb_career",
        selectable=select(
            [
                db.PitchFx.batter_id.label("id"),
                db.PitchFx.batter_id_mlb.label("mlb_id"),
                db.PitchFx.stand.label("bat_stand"),
                db.PitchFx.p_throws.label("p_throws"),
                literal("ALL").label("pitch_type"),
                func.count(db.PitchFx.id).label("total_pitches"),
                func.sum(db.PitchFx.is_final_pitch_of_ab).label("total_pa"),
                total_at_bats,
                func.sum(db.PitchFx.ab_result_out).label("total_outs"),
                func.sum(db.PitchFx.ab_result_hit).label("total_hits"),
                func.sum(db.PitchFx.ab_result_bb).label("total_bb"),
                func.sum(db.PitchFx.ab_result_k).label("total_k"),
                avg,
                obp,
                slg,
                ops,
                iso,
                fly_ball_rate,
                ground_ball_rate,
                line_drive_rate,
                pop_up_rate,
                bb_rate,
                k_rate,
                hr_per_fb,
                zone_rate,
                called_strike_rate,
                swinging_strike_rate,
                whiff_rate,
                csw_rate,
                o_swing_rate,
                z_swing_rate,
                swing_rate,
                o_contact_rate,
                z_contact_rate,
                contact_rate,
                func.sum(db.PitchFx.batter_did_swing).label("total_swings"),
                func.sum(db.PitchFx.batter_made_contact).label("total_swings_made_contact"),
                func.sum(db.PitchFx.called_strike).label("total_called_strikes"),
                func.sum(db.PitchFx.swinging_strike).label("total_swinging_strikes"),
                func.sum(db.PitchFx.inside_strike_zone).label("total_inside_strike_zone"),
                func.sum(db.PitchFx.outside_strike_zone).label("total_outside_strike_zone"),
                func.sum(db.PitchFx.swing_inside_zone).label("total_swings_inside_zone"),
                func.sum(db.PitchFx.swing_outside_zone).label("total_swings_outside_zone"),
                func.sum(db.PitchFx.contact_inside_zone).label("total_contact_inside_zone"),
                func.sum(db.PitchFx.contact_outside_zone).label("total_contact_outside_zone"),
                func.sum(db.PitchFx.is_batted_ball).label("total_batted_balls"),
                func.sum(db.PitchFx.is_ground_ball).label("total_ground_balls"),
                func.sum(db.PitchFx.is_line_drive).label("total_line_drives"),
                func.sum(db.PitchFx.is_fly_ball).label("total_fly_balls"),
                func.sum(db.PitchFx.is_pop_up).label("total_pop_ups"),
                func.sum(db.PitchFx.ab_result_single).label("total_singles"),
                func.sum(db.PitchFx.ab_result_double).label("total_doubles"),
                func.sum(db.PitchFx.ab_result_triple).label("total_triples"),
                func.sum(db.PitchFx.ab_result_homerun).label("total_homeruns"),
                func.sum(db.PitchFx.ab_result_ibb).label("total_ibb"),
                func.sum(db.PitchFx.ab_result_hbp).label("total_hbp"),
                func.sum(db.PitchFx.ab_result_error).label("total_errors"),
                func.sum(db.PitchFx.ab_result_sac_hit).label("total_sac_hit"),
                func.sum(db.PitchFx.ab_result_sac_fly).label("total_sac_fly"),
                func.sum(db.PitchFx.ab_result_unclear).label("total_unclear_results"),
                func.sum(db.PitchFx.pitch_type_int.distinct()).label("pitch_type_int"),
            ]
        )
        .where(
            and_(
                db.PitchFx.is_invalid_ibb == 0,
                db.PitchFx.is_out_of_sequence == 0,
                db.PitchFx.stand == "R",
                db.PitchFx.p_throws == "L",
            )
        )
        .select_from(db.PitchFx)
        .group_by(db.PitchFx.batter_id)
        .order_by(db.PitchFx.batter_id_mlb),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pfx_metrics_vs_lhp_as_rhb_for_career_for_batter(cls, db_engine, mlb_id):
        results = db_engine.execute(select([cls]).where(cls.mlb_id == mlb_id)).fetchall()
        return [dict(row) for row in results][0] if results else {}


class Batter_PitchType_vs_LHP_as_RHB_Career_View(db.Base):
    __table__ = create_view(
        name="batter_pitchtype_vs_lhp_as_rhb_career",
        selectable=select(
            [
                db.PitchFx.batter_id.label("id"),
                db.PitchFx.batter_id_mlb.label("mlb_id"),
                db.PitchFx.stand.label("bat_stand"),
                db.PitchFx.p_throws.label("p_throws"),
                db.PitchFx.mlbam_pitch_name.label("pitch_type"),
                func.count(db.PitchFx.id).label("total_pitches"),
                func.sum(db.PitchFx.is_final_pitch_of_ab).label("total_pa"),
                total_at_bats,
                func.sum(db.PitchFx.ab_result_out).label("total_outs"),
                func.sum(db.PitchFx.ab_result_hit).label("total_hits"),
                func.sum(db.PitchFx.ab_result_bb).label("total_bb"),
                func.sum(db.PitchFx.ab_result_k).label("total_k"),
                avg,
                obp,
                slg,
                ops,
                iso,
                fly_ball_rate,
                ground_ball_rate,
                line_drive_rate,
                pop_up_rate,
                bb_rate,
                k_rate,
                hr_per_fb,
                zone_rate,
                called_strike_rate,
                swinging_strike_rate,
                whiff_rate,
                csw_rate,
                o_swing_rate,
                z_swing_rate,
                swing_rate,
                o_contact_rate,
                z_contact_rate,
                contact_rate,
                func.sum(db.PitchFx.batter_did_swing).label("total_swings"),
                func.sum(db.PitchFx.batter_made_contact).label("total_swings_made_contact"),
                func.sum(db.PitchFx.called_strike).label("total_called_strikes"),
                func.sum(db.PitchFx.swinging_strike).label("total_swinging_strikes"),
                func.sum(db.PitchFx.inside_strike_zone).label("total_inside_strike_zone"),
                func.sum(db.PitchFx.outside_strike_zone).label("total_outside_strike_zone"),
                func.sum(db.PitchFx.swing_inside_zone).label("total_swings_inside_zone"),
                func.sum(db.PitchFx.swing_outside_zone).label("total_swings_outside_zone"),
                func.sum(db.PitchFx.contact_inside_zone).label("total_contact_inside_zone"),
                func.sum(db.PitchFx.contact_outside_zone).label("total_contact_outside_zone"),
                func.sum(db.PitchFx.is_batted_ball).label("total_batted_balls"),
                func.sum(db.PitchFx.is_ground_ball).label("total_ground_balls"),
                func.sum(db.PitchFx.is_line_drive).label("total_line_drives"),
                func.sum(db.PitchFx.is_fly_ball).label("total_fly_balls"),
                func.sum(db.PitchFx.is_pop_up).label("total_pop_ups"),
                func.sum(db.PitchFx.ab_result_single).label("total_singles"),
                func.sum(db.PitchFx.ab_result_double).label("total_doubles"),
                func.sum(db.PitchFx.ab_result_triple).label("total_triples"),
                func.sum(db.PitchFx.ab_result_homerun).label("total_homeruns"),
                func.sum(db.PitchFx.ab_result_ibb).label("total_ibb"),
                func.sum(db.PitchFx.ab_result_hbp).label("total_hbp"),
                func.sum(db.PitchFx.ab_result_error).label("total_errors"),
                func.sum(db.PitchFx.ab_result_sac_hit).label("total_sac_hit"),
                func.sum(db.PitchFx.ab_result_sac_fly).label("total_sac_fly"),
                func.sum(db.PitchFx.ab_result_unclear).label("total_unclear_results"),
                func.sum(db.PitchFx.pitch_type_int.distinct()).label("pitch_type_int"),
            ]
        )
        .where(
            and_(
                db.PitchFx.is_invalid_ibb == 0,
                db.PitchFx.is_out_of_sequence == 0,
                db.PitchFx.stand == "R",
                db.PitchFx.p_throws == "L",
            )
        )
        .select_from(db.PitchFx)
        .group_by(db.PitchFx.batter_id)
        .group_by(db.PitchFx.mlbam_pitch_name)
        .order_by(db.PitchFx.batter_id_mlb),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pfx_metrics_vs_lhp_as_rhb_for_career_for_batter(cls, db_engine, mlb_id):
        results = db_engine.execute(select([cls]).where(cls.mlb_id == mlb_id)).fetchall()
        return [dict(row) for row in results] if results else []


class Batter_PitchFx_By_Year_View(db.Base):
    __table__ = create_view(
        name="batter_pitchfx_by_year",
        selectable=select(
            [
                db.PitchFx.batter_id.label("id"),
                db.PitchFx.batter_id_mlb.label("mlb_id"),
                db.Season.year.label("year"),
                literal("ALL").label("pitch_type"),
                func.count(db.PitchFx.id).label("total_pitches"),
                func.sum(db.PitchFx.is_final_pitch_of_ab).label("total_pa"),
                total_at_bats,
                func.sum(db.PitchFx.ab_result_out).label("total_outs"),
                func.sum(db.PitchFx.ab_result_hit).label("total_hits"),
                func.sum(db.PitchFx.ab_result_bb).label("total_bb"),
                func.sum(db.PitchFx.ab_result_k).label("total_k"),
                avg,
                obp,
                slg,
                ops,
                iso,
                fly_ball_rate,
                ground_ball_rate,
                line_drive_rate,
                pop_up_rate,
                bb_rate,
                k_rate,
                hr_per_fb,
                zone_rate,
                called_strike_rate,
                swinging_strike_rate,
                whiff_rate,
                csw_rate,
                o_swing_rate,
                z_swing_rate,
                swing_rate,
                o_contact_rate,
                z_contact_rate,
                contact_rate,
                func.sum(db.PitchFx.batter_did_swing).label("total_swings"),
                func.sum(db.PitchFx.batter_made_contact).label("total_swings_made_contact"),
                func.sum(db.PitchFx.called_strike).label("total_called_strikes"),
                func.sum(db.PitchFx.swinging_strike).label("total_swinging_strikes"),
                func.sum(db.PitchFx.inside_strike_zone).label("total_inside_strike_zone"),
                func.sum(db.PitchFx.outside_strike_zone).label("total_outside_strike_zone"),
                func.sum(db.PitchFx.swing_inside_zone).label("total_swings_inside_zone"),
                func.sum(db.PitchFx.swing_outside_zone).label("total_swings_outside_zone"),
                func.sum(db.PitchFx.contact_inside_zone).label("total_contact_inside_zone"),
                func.sum(db.PitchFx.contact_outside_zone).label("total_contact_outside_zone"),
                func.sum(db.PitchFx.is_batted_ball).label("total_batted_balls"),
                func.sum(db.PitchFx.is_ground_ball).label("total_ground_balls"),
                func.sum(db.PitchFx.is_line_drive).label("total_line_drives"),
                func.sum(db.PitchFx.is_fly_ball).label("total_fly_balls"),
                func.sum(db.PitchFx.is_pop_up).label("total_pop_ups"),
                func.sum(db.PitchFx.ab_result_single).label("total_singles"),
                func.sum(db.PitchFx.ab_result_double).label("total_doubles"),
                func.sum(db.PitchFx.ab_result_triple).label("total_triples"),
                func.sum(db.PitchFx.ab_result_homerun).label("total_homeruns"),
                func.sum(db.PitchFx.ab_result_ibb).label("total_ibb"),
                func.sum(db.PitchFx.ab_result_hbp).label("total_hbp"),
                func.sum(db.PitchFx.ab_result_error).label("total_errors"),
                func.sum(db.PitchFx.ab_result_sac_hit).label("total_sac_hit"),
                func.sum(db.PitchFx.ab_result_sac_fly).label("total_sac_fly"),
                func.sum(db.PitchFx.ab_result_unclear).label("total_unclear_results"),
                func.sum(db.PitchFx.pitch_type_int.distinct()).label("pitch_type_int"),
                db.PitchFx.season_id.label("season_id"),
            ]
        )
        .where(
            and_(
                db.PitchFx.is_invalid_ibb == 0,
                db.PitchFx.is_out_of_sequence == 0,
                or_(db.PitchFx.stand == "L", db.PitchFx.stand == "R"),
            )
        )
        .select_from(join(db.PitchFx, db.Season, db.PitchFx.season_id == db.Season.id))
        .group_by(db.PitchFx.season_id)
        .group_by(db.PitchFx.batter_id)
        .order_by(db.PitchFx.batter_id_mlb),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pfx_metrics_by_year_for_batter(cls, db_engine, mlb_id, year):
        s = select([cls]).where(and_(cls.mlb_id == mlb_id, cls.year == year))
        results = db_engine.execute(s).fetchall()
        return [dict(row) for row in results][0] if results else {}


class Batter_PitchType_By_Year_View(db.Base):
    __table__ = create_view(
        name="batter_pitchtype_by_year",
        selectable=select(
            [
                db.PitchFx.batter_id.label("id"),
                db.PitchFx.batter_id_mlb.label("mlb_id"),
                db.Season.year.label("year"),
                db.PitchFx.mlbam_pitch_name.label("pitch_type"),
                func.count(db.PitchFx.id).label("total_pitches"),
                func.sum(db.PitchFx.is_final_pitch_of_ab).label("total_pa"),
                total_at_bats,
                func.sum(db.PitchFx.ab_result_out).label("total_outs"),
                func.sum(db.PitchFx.ab_result_hit).label("total_hits"),
                func.sum(db.PitchFx.ab_result_bb).label("total_bb"),
                func.sum(db.PitchFx.ab_result_k).label("total_k"),
                func.avg(db.PitchFx.start_speed).label("avg_speed"),
                func.avg(db.PitchFx.pfx_x).label("avg_pfx_x"),
                func.avg(db.PitchFx.pfx_z).label("avg_pfx_z"),
                func.avg(db.PitchFx.px).label("avg_px"),
                func.avg(db.PitchFx.pz).label("avg_pz"),
                avg,
                obp,
                slg,
                ops,
                iso,
                fly_ball_rate,
                ground_ball_rate,
                line_drive_rate,
                pop_up_rate,
                bb_rate,
                k_rate,
                hr_per_fb,
                zone_rate,
                called_strike_rate,
                swinging_strike_rate,
                whiff_rate,
                csw_rate,
                o_swing_rate,
                z_swing_rate,
                swing_rate,
                o_contact_rate,
                z_contact_rate,
                contact_rate,
                func.sum(db.PitchFx.batter_did_swing).label("total_swings"),
                func.sum(db.PitchFx.batter_made_contact).label("total_swings_made_contact"),
                func.sum(db.PitchFx.called_strike).label("total_called_strikes"),
                func.sum(db.PitchFx.swinging_strike).label("total_swinging_strikes"),
                func.sum(db.PitchFx.inside_strike_zone).label("total_inside_strike_zone"),
                func.sum(db.PitchFx.outside_strike_zone).label("total_outside_strike_zone"),
                func.sum(db.PitchFx.swing_inside_zone).label("total_swings_inside_zone"),
                func.sum(db.PitchFx.swing_outside_zone).label("total_swings_outside_zone"),
                func.sum(db.PitchFx.contact_inside_zone).label("total_contact_inside_zone"),
                func.sum(db.PitchFx.contact_outside_zone).label("total_contact_outside_zone"),
                func.sum(db.PitchFx.is_batted_ball).label("total_batted_balls"),
                func.sum(db.PitchFx.is_ground_ball).label("total_ground_balls"),
                func.sum(db.PitchFx.is_line_drive).label("total_line_drives"),
                func.sum(db.PitchFx.is_fly_ball).label("total_fly_balls"),
                func.sum(db.PitchFx.is_pop_up).label("total_pop_ups"),
                func.sum(db.PitchFx.ab_result_single).label("total_singles"),
                func.sum(db.PitchFx.ab_result_double).label("total_doubles"),
                func.sum(db.PitchFx.ab_result_triple).label("total_triples"),
                func.sum(db.PitchFx.ab_result_homerun).label("total_homeruns"),
                func.sum(db.PitchFx.ab_result_ibb).label("total_ibb"),
                func.sum(db.PitchFx.ab_result_hbp).label("total_hbp"),
                func.sum(db.PitchFx.ab_result_error).label("total_errors"),
                func.sum(db.PitchFx.ab_result_sac_hit).label("total_sac_hit"),
                func.sum(db.PitchFx.ab_result_sac_fly).label("total_sac_fly"),
                func.sum(db.PitchFx.ab_result_unclear).label("total_unclear_results"),
                db.PitchFx.season_id.label("season_id"),
            ]
        )
        .where(
            and_(
                db.PitchFx.is_invalid_ibb == 0,
                db.PitchFx.is_out_of_sequence == 0,
                or_(db.PitchFx.stand == "L", db.PitchFx.stand == "R"),
            )
        )
        .select_from(join(db.PitchFx, db.Season, db.PitchFx.season_id == db.Season.id))
        .group_by(db.PitchFx.season_id)
        .group_by(db.PitchFx.batter_id)
        .group_by(db.PitchFx.mlbam_pitch_name)
        .order_by(db.PitchFx.batter_id_mlb),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pfx_metrics_by_year_for_batter(cls, db_engine, mlb_id, year):
        s = select([cls]).where(and_(cls.mlb_id == mlb_id, cls.year == year))
        results = db_engine.execute(s).fetchall()
        return [dict(row) for row in results] if results else []


class Batter_PitchFx_For_Game_All_View(db.Base):
    __table__ = create_view(
        name="batter_pitchfx_for_game_all",
        selectable=select(
            [
                db.PitchFx.batter_id.label("id"),
                db.PitchFx.batter_id_mlb.label("mlb_id"),
                db.PitchFx.bbref_game_id.label("bbref_game_id"),
                literal("ALL").label("pitch_type"),
                func.count(db.PitchFx.id).label("total_pitches"),
                func.sum(db.PitchFx.is_final_pitch_of_ab).label("total_pa"),
                total_at_bats,
                func.sum(db.PitchFx.ab_result_out).label("total_outs"),
                func.sum(db.PitchFx.ab_result_hit).label("total_hits"),
                func.sum(db.PitchFx.ab_result_bb).label("total_bb"),
                func.sum(db.PitchFx.ab_result_k).label("total_k"),
                avg,
                obp,
                slg,
                ops,
                iso,
                fly_ball_rate,
                ground_ball_rate,
                line_drive_rate,
                pop_up_rate,
                bb_rate,
                k_rate,
                hr_per_fb,
                zone_rate,
                called_strike_rate,
                swinging_strike_rate,
                whiff_rate,
                csw_rate,
                o_swing_rate,
                z_swing_rate,
                swing_rate,
                o_contact_rate,
                z_contact_rate,
                contact_rate,
                func.sum(db.PitchFx.batter_did_swing).label("total_swings"),
                func.sum(db.PitchFx.batter_made_contact).label("total_swings_made_contact"),
                func.sum(db.PitchFx.called_strike).label("total_called_strikes"),
                func.sum(db.PitchFx.swinging_strike).label("total_swinging_strikes"),
                func.sum(db.PitchFx.inside_strike_zone).label("total_inside_strike_zone"),
                func.sum(db.PitchFx.outside_strike_zone).label("total_outside_strike_zone"),
                func.sum(db.PitchFx.swing_inside_zone).label("total_swings_inside_zone"),
                func.sum(db.PitchFx.swing_outside_zone).label("total_swings_outside_zone"),
                func.sum(db.PitchFx.contact_inside_zone).label("total_contact_inside_zone"),
                func.sum(db.PitchFx.contact_outside_zone).label("total_contact_outside_zone"),
                func.sum(db.PitchFx.is_batted_ball).label("total_batted_balls"),
                func.sum(db.PitchFx.is_ground_ball).label("total_ground_balls"),
                func.sum(db.PitchFx.is_line_drive).label("total_line_drives"),
                func.sum(db.PitchFx.is_fly_ball).label("total_fly_balls"),
                func.sum(db.PitchFx.is_pop_up).label("total_pop_ups"),
                func.sum(db.PitchFx.ab_result_single).label("total_singles"),
                func.sum(db.PitchFx.ab_result_double).label("total_doubles"),
                func.sum(db.PitchFx.ab_result_triple).label("total_triples"),
                func.sum(db.PitchFx.ab_result_homerun).label("total_homeruns"),
                func.sum(db.PitchFx.ab_result_ibb).label("total_ibb"),
                func.sum(db.PitchFx.ab_result_hbp).label("total_hbp"),
                func.sum(db.PitchFx.ab_result_error).label("total_errors"),
                func.sum(db.PitchFx.ab_result_sac_hit).label("total_sac_hit"),
                func.sum(db.PitchFx.ab_result_sac_fly).label("total_sac_fly"),
                func.sum(db.PitchFx.ab_result_unclear).label("total_unclear_results"),
                func.sum(db.PitchFx.pitch_type_int.distinct()).label("pitch_type_int"),
                db.PitchFx.team_pitching_id.label("team_pitching_id"),
                db.PitchFx.team_batting_id.label("team_batting_id"),
                db.PitchFx.game_status_id.label("game_status_id"),
                db.PitchFx.date_id.label("date_id"),
                db.PitchFx.season_id.label("season_id"),
            ]
        )
        .where(
            and_(
                db.PitchFx.is_invalid_ibb == 0,
                db.PitchFx.is_out_of_sequence == 0,
                or_(db.PitchFx.stand == "L", db.PitchFx.stand == "R"),
                or_(db.PitchFx.p_throws == "L", db.PitchFx.p_throws == "R"),
            )
        )
        .select_from(db.PitchFx)
        .group_by(db.PitchFx.batter_id)
        .group_by(db.PitchFx.game_status_id)
        .order_by(db.PitchFx.batter_id_mlb)
        .order_by(db.PitchFx.date_id),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pfx_metrics_for_game_for_batter(cls, db_engine, mlb_id, game_id):
        s = select([cls]).where(and_(cls.mlb_id == mlb_id, cls.bbref_game_id == game_id))
        results = db_engine.execute(s).fetchall()
        return [dict(row) for row in results][0] if results else {}


class Batter_PitchType_For_Game_All_View(db.Base):
    __table__ = create_view(
        name="batter_pitchtype_for_game_all",
        selectable=select(
            [
                db.PitchFx.batter_id.label("id"),
                db.PitchFx.batter_id_mlb.label("mlb_id"),
                db.PitchFx.bbref_game_id.label("bbref_game_id"),
                db.PitchFx.mlbam_pitch_name.label("pitch_type"),
                func.count(db.PitchFx.id).label("total_pitches"),
                func.sum(db.PitchFx.is_final_pitch_of_ab).label("total_pa"),
                total_at_bats,
                func.sum(db.PitchFx.ab_result_out).label("total_outs"),
                func.sum(db.PitchFx.ab_result_hit).label("total_hits"),
                func.sum(db.PitchFx.ab_result_bb).label("total_bb"),
                func.sum(db.PitchFx.ab_result_k).label("total_k"),
                avg,
                obp,
                slg,
                ops,
                iso,
                fly_ball_rate,
                ground_ball_rate,
                line_drive_rate,
                pop_up_rate,
                bb_rate,
                k_rate,
                hr_per_fb,
                zone_rate,
                called_strike_rate,
                swinging_strike_rate,
                whiff_rate,
                csw_rate,
                o_swing_rate,
                z_swing_rate,
                swing_rate,
                o_contact_rate,
                z_contact_rate,
                contact_rate,
                func.sum(db.PitchFx.batter_did_swing).label("total_swings"),
                func.sum(db.PitchFx.batter_made_contact).label("total_swings_made_contact"),
                func.sum(db.PitchFx.called_strike).label("total_called_strikes"),
                func.sum(db.PitchFx.swinging_strike).label("total_swinging_strikes"),
                func.sum(db.PitchFx.inside_strike_zone).label("total_inside_strike_zone"),
                func.sum(db.PitchFx.outside_strike_zone).label("total_outside_strike_zone"),
                func.sum(db.PitchFx.swing_inside_zone).label("total_swings_inside_zone"),
                func.sum(db.PitchFx.swing_outside_zone).label("total_swings_outside_zone"),
                func.sum(db.PitchFx.contact_inside_zone).label("total_contact_inside_zone"),
                func.sum(db.PitchFx.contact_outside_zone).label("total_contact_outside_zone"),
                func.sum(db.PitchFx.is_batted_ball).label("total_batted_balls"),
                func.sum(db.PitchFx.is_ground_ball).label("total_ground_balls"),
                func.sum(db.PitchFx.is_line_drive).label("total_line_drives"),
                func.sum(db.PitchFx.is_fly_ball).label("total_fly_balls"),
                func.sum(db.PitchFx.is_pop_up).label("total_pop_ups"),
                func.sum(db.PitchFx.ab_result_single).label("total_singles"),
                func.sum(db.PitchFx.ab_result_double).label("total_doubles"),
                func.sum(db.PitchFx.ab_result_triple).label("total_triples"),
                func.sum(db.PitchFx.ab_result_homerun).label("total_homeruns"),
                func.sum(db.PitchFx.ab_result_ibb).label("total_ibb"),
                func.sum(db.PitchFx.ab_result_hbp).label("total_hbp"),
                func.sum(db.PitchFx.ab_result_error).label("total_errors"),
                func.sum(db.PitchFx.ab_result_sac_hit).label("total_sac_hit"),
                func.sum(db.PitchFx.ab_result_sac_fly).label("total_sac_fly"),
                func.sum(db.PitchFx.ab_result_unclear).label("total_unclear_results"),
                func.sum(db.PitchFx.pitch_type_int.distinct()).label("pitch_type_int"),
                db.PitchFx.team_pitching_id.label("team_pitching_id"),
                db.PitchFx.team_batting_id.label("team_batting_id"),
                db.PitchFx.game_status_id.label("game_status_id"),
                db.PitchFx.date_id.label("date_id"),
                db.PitchFx.season_id.label("season_id"),
            ]
        )
        .where(
            and_(
                db.PitchFx.is_invalid_ibb == 0,
                db.PitchFx.is_out_of_sequence == 0,
                or_(db.PitchFx.stand == "L", db.PitchFx.stand == "R"),
                or_(db.PitchFx.p_throws == "L", db.PitchFx.p_throws == "R"),
            )
        )
        .select_from(db.PitchFx)
        .group_by(db.PitchFx.batter_id)
        .group_by(db.PitchFx.game_status_id)
        .group_by(db.PitchFx.mlbam_pitch_name)
        .order_by(db.PitchFx.batter_id_mlb)
        .order_by(db.PitchFx.date_id),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pfx_metrics_for_game_for_batter(cls, db_engine, mlb_id, game_id):
        s = select([cls]).where(and_(cls.mlb_id == mlb_id, cls.bbref_game_id == game_id))
        results = db_engine.execute(s).fetchall()
        return [dict(row) for row in results] if results else []


class Batter_PitchFx_For_Game_vs_RHP_as_RHB_View(db.Base):
    __table__ = create_view(
        name="batter_pitchfx_vs_rhp_as_rhb_for_game",
        selectable=select(
            [
                db.PitchFx.batter_id.label("id"),
                db.PitchFx.batter_id_mlb.label("mlb_id"),
                db.PitchFx.bbref_game_id.label("bbref_game_id"),
                db.PitchFx.stand.label("bat_stand"),
                db.PitchFx.p_throws.label("p_throws"),
                literal("ALL").label("pitch_type"),
                func.count(db.PitchFx.id).label("total_pitches"),
                func.sum(db.PitchFx.is_final_pitch_of_ab).label("total_pa"),
                total_at_bats,
                func.sum(db.PitchFx.ab_result_out).label("total_outs"),
                func.sum(db.PitchFx.ab_result_hit).label("total_hits"),
                func.sum(db.PitchFx.ab_result_bb).label("total_bb"),
                func.sum(db.PitchFx.ab_result_k).label("total_k"),
                avg,
                obp,
                slg,
                ops,
                iso,
                fly_ball_rate,
                ground_ball_rate,
                line_drive_rate,
                pop_up_rate,
                bb_rate,
                k_rate,
                hr_per_fb,
                zone_rate,
                called_strike_rate,
                swinging_strike_rate,
                whiff_rate,
                csw_rate,
                o_swing_rate,
                z_swing_rate,
                swing_rate,
                o_contact_rate,
                z_contact_rate,
                contact_rate,
                func.sum(db.PitchFx.batter_did_swing).label("total_swings"),
                func.sum(db.PitchFx.batter_made_contact).label("total_swings_made_contact"),
                func.sum(db.PitchFx.called_strike).label("total_called_strikes"),
                func.sum(db.PitchFx.swinging_strike).label("total_swinging_strikes"),
                func.sum(db.PitchFx.inside_strike_zone).label("total_inside_strike_zone"),
                func.sum(db.PitchFx.outside_strike_zone).label("total_outside_strike_zone"),
                func.sum(db.PitchFx.swing_inside_zone).label("total_swings_inside_zone"),
                func.sum(db.PitchFx.swing_outside_zone).label("total_swings_outside_zone"),
                func.sum(db.PitchFx.contact_inside_zone).label("total_contact_inside_zone"),
                func.sum(db.PitchFx.contact_outside_zone).label("total_contact_outside_zone"),
                func.sum(db.PitchFx.is_batted_ball).label("total_batted_balls"),
                func.sum(db.PitchFx.is_ground_ball).label("total_ground_balls"),
                func.sum(db.PitchFx.is_line_drive).label("total_line_drives"),
                func.sum(db.PitchFx.is_fly_ball).label("total_fly_balls"),
                func.sum(db.PitchFx.is_pop_up).label("total_pop_ups"),
                func.sum(db.PitchFx.ab_result_single).label("total_singles"),
                func.sum(db.PitchFx.ab_result_double).label("total_doubles"),
                func.sum(db.PitchFx.ab_result_triple).label("total_triples"),
                func.sum(db.PitchFx.ab_result_homerun).label("total_homeruns"),
                func.sum(db.PitchFx.ab_result_ibb).label("total_ibb"),
                func.sum(db.PitchFx.ab_result_hbp).label("total_hbp"),
                func.sum(db.PitchFx.ab_result_error).label("total_errors"),
                func.sum(db.PitchFx.ab_result_sac_hit).label("total_sac_hit"),
                func.sum(db.PitchFx.ab_result_sac_fly).label("total_sac_fly"),
                func.sum(db.PitchFx.ab_result_unclear).label("total_unclear_results"),
                func.sum(db.PitchFx.pitch_type_int.distinct()).label("pitch_type_int"),
                db.PitchFx.team_pitching_id.label("team_pitching_id"),
                db.PitchFx.team_batting_id.label("team_batting_id"),
                db.PitchFx.game_status_id.label("game_status_id"),
                db.PitchFx.date_id.label("date_id"),
                db.PitchFx.season_id.label("season_id"),
            ]
        )
        .where(
            and_(
                db.PitchFx.is_invalid_ibb == 0,
                db.PitchFx.is_out_of_sequence == 0,
                db.PitchFx.stand == "R",
                db.PitchFx.p_throws == "R",
            )
        )
        .select_from(db.PitchFx)
        .group_by(db.PitchFx.batter_id)
        .group_by(db.PitchFx.game_status_id)
        .order_by(db.PitchFx.batter_id_mlb)
        .order_by(db.PitchFx.date_id),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pfx_metrics_vs_rhp_as_rhb_for_game_for_batter(cls, db_engine, mlb_id, game_id):
        s = select([cls]).where(and_(cls.mlb_id == mlb_id, cls.bbref_game_id == game_id))
        results = db_engine.execute(s).fetchall()
        return [dict(row) for row in results][0] if results else {}


class Batter_PitchType_For_Game_vs_RHP_as_RHB_View(db.Base):
    __table__ = create_view(
        name="batter_pitchtype_vs_rhp_as_rhb_for_game",
        selectable=select(
            [
                db.PitchFx.batter_id.label("id"),
                db.PitchFx.batter_id_mlb.label("mlb_id"),
                db.PitchFx.bbref_game_id.label("bbref_game_id"),
                db.PitchFx.stand.label("bat_stand"),
                db.PitchFx.p_throws.label("p_throws"),
                db.PitchFx.mlbam_pitch_name.label("pitch_type"),
                func.count(db.PitchFx.id).label("total_pitches"),
                func.sum(db.PitchFx.is_final_pitch_of_ab).label("total_pa"),
                total_at_bats,
                func.sum(db.PitchFx.ab_result_out).label("total_outs"),
                func.sum(db.PitchFx.ab_result_hit).label("total_hits"),
                func.sum(db.PitchFx.ab_result_bb).label("total_bb"),
                func.sum(db.PitchFx.ab_result_k).label("total_k"),
                avg,
                obp,
                slg,
                ops,
                iso,
                fly_ball_rate,
                ground_ball_rate,
                line_drive_rate,
                pop_up_rate,
                bb_rate,
                k_rate,
                hr_per_fb,
                zone_rate,
                called_strike_rate,
                swinging_strike_rate,
                whiff_rate,
                csw_rate,
                o_swing_rate,
                z_swing_rate,
                swing_rate,
                o_contact_rate,
                z_contact_rate,
                contact_rate,
                func.sum(db.PitchFx.batter_did_swing).label("total_swings"),
                func.sum(db.PitchFx.batter_made_contact).label("total_swings_made_contact"),
                func.sum(db.PitchFx.called_strike).label("total_called_strikes"),
                func.sum(db.PitchFx.swinging_strike).label("total_swinging_strikes"),
                func.sum(db.PitchFx.inside_strike_zone).label("total_inside_strike_zone"),
                func.sum(db.PitchFx.outside_strike_zone).label("total_outside_strike_zone"),
                func.sum(db.PitchFx.swing_inside_zone).label("total_swings_inside_zone"),
                func.sum(db.PitchFx.swing_outside_zone).label("total_swings_outside_zone"),
                func.sum(db.PitchFx.contact_inside_zone).label("total_contact_inside_zone"),
                func.sum(db.PitchFx.contact_outside_zone).label("total_contact_outside_zone"),
                func.sum(db.PitchFx.is_batted_ball).label("total_batted_balls"),
                func.sum(db.PitchFx.is_ground_ball).label("total_ground_balls"),
                func.sum(db.PitchFx.is_line_drive).label("total_line_drives"),
                func.sum(db.PitchFx.is_fly_ball).label("total_fly_balls"),
                func.sum(db.PitchFx.is_pop_up).label("total_pop_ups"),
                func.sum(db.PitchFx.ab_result_single).label("total_singles"),
                func.sum(db.PitchFx.ab_result_double).label("total_doubles"),
                func.sum(db.PitchFx.ab_result_triple).label("total_triples"),
                func.sum(db.PitchFx.ab_result_homerun).label("total_homeruns"),
                func.sum(db.PitchFx.ab_result_ibb).label("total_ibb"),
                func.sum(db.PitchFx.ab_result_hbp).label("total_hbp"),
                func.sum(db.PitchFx.ab_result_error).label("total_errors"),
                func.sum(db.PitchFx.ab_result_sac_hit).label("total_sac_hit"),
                func.sum(db.PitchFx.ab_result_sac_fly).label("total_sac_fly"),
                func.sum(db.PitchFx.ab_result_unclear).label("total_unclear_results"),
                func.sum(db.PitchFx.pitch_type_int.distinct()).label("pitch_type_int"),
                db.PitchFx.team_pitching_id.label("team_pitching_id"),
                db.PitchFx.team_batting_id.label("team_batting_id"),
                db.PitchFx.game_status_id.label("game_status_id"),
                db.PitchFx.date_id.label("date_id"),
                db.PitchFx.season_id.label("season_id"),
            ]
        )
        .where(
            and_(
                db.PitchFx.is_invalid_ibb == 0,
                db.PitchFx.is_out_of_sequence == 0,
                db.PitchFx.stand == "R",
                db.PitchFx.p_throws == "R",
            )
        )
        .select_from(db.PitchFx)
        .group_by(db.PitchFx.batter_id)
        .group_by(db.PitchFx.game_status_id)
        .group_by(db.PitchFx.mlbam_pitch_name)
        .order_by(db.PitchFx.batter_id_mlb)
        .order_by(db.PitchFx.date_id),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pfx_metrics_vs_rhp_as_rhb_for_game_for_batter(cls, db_engine, mlb_id, game_id):
        s = select([cls]).where(and_(cls.mlb_id == mlb_id, cls.bbref_game_id == game_id))
        results = db_engine.execute(s).fetchall()
        return [dict(row) for row in results] if results else []


class Batter_PitchFx_For_Game_vs_RHP_as_LHB_View(db.Base):
    __table__ = create_view(
        name="batter_pitchfx_vs_rhp_as_lhb_for_game",
        selectable=select(
            [
                db.PitchFx.batter_id.label("id"),
                db.PitchFx.batter_id_mlb.label("mlb_id"),
                db.PitchFx.bbref_game_id.label("bbref_game_id"),
                db.PitchFx.stand.label("bat_stand"),
                db.PitchFx.p_throws.label("p_throws"),
                literal("ALL").label("pitch_type"),
                func.count(db.PitchFx.id).label("total_pitches"),
                func.sum(db.PitchFx.is_final_pitch_of_ab).label("total_pa"),
                total_at_bats,
                func.sum(db.PitchFx.ab_result_out).label("total_outs"),
                func.sum(db.PitchFx.ab_result_hit).label("total_hits"),
                func.sum(db.PitchFx.ab_result_bb).label("total_bb"),
                func.sum(db.PitchFx.ab_result_k).label("total_k"),
                avg,
                obp,
                slg,
                ops,
                iso,
                fly_ball_rate,
                ground_ball_rate,
                line_drive_rate,
                pop_up_rate,
                bb_rate,
                k_rate,
                hr_per_fb,
                zone_rate,
                called_strike_rate,
                swinging_strike_rate,
                whiff_rate,
                csw_rate,
                o_swing_rate,
                z_swing_rate,
                swing_rate,
                o_contact_rate,
                z_contact_rate,
                contact_rate,
                func.sum(db.PitchFx.batter_did_swing).label("total_swings"),
                func.sum(db.PitchFx.batter_made_contact).label("total_swings_made_contact"),
                func.sum(db.PitchFx.called_strike).label("total_called_strikes"),
                func.sum(db.PitchFx.swinging_strike).label("total_swinging_strikes"),
                func.sum(db.PitchFx.inside_strike_zone).label("total_inside_strike_zone"),
                func.sum(db.PitchFx.outside_strike_zone).label("total_outside_strike_zone"),
                func.sum(db.PitchFx.swing_inside_zone).label("total_swings_inside_zone"),
                func.sum(db.PitchFx.swing_outside_zone).label("total_swings_outside_zone"),
                func.sum(db.PitchFx.contact_inside_zone).label("total_contact_inside_zone"),
                func.sum(db.PitchFx.contact_outside_zone).label("total_contact_outside_zone"),
                func.sum(db.PitchFx.is_batted_ball).label("total_batted_balls"),
                func.sum(db.PitchFx.is_ground_ball).label("total_ground_balls"),
                func.sum(db.PitchFx.is_line_drive).label("total_line_drives"),
                func.sum(db.PitchFx.is_fly_ball).label("total_fly_balls"),
                func.sum(db.PitchFx.is_pop_up).label("total_pop_ups"),
                func.sum(db.PitchFx.ab_result_single).label("total_singles"),
                func.sum(db.PitchFx.ab_result_double).label("total_doubles"),
                func.sum(db.PitchFx.ab_result_triple).label("total_triples"),
                func.sum(db.PitchFx.ab_result_homerun).label("total_homeruns"),
                func.sum(db.PitchFx.ab_result_ibb).label("total_ibb"),
                func.sum(db.PitchFx.ab_result_hbp).label("total_hbp"),
                func.sum(db.PitchFx.ab_result_error).label("total_errors"),
                func.sum(db.PitchFx.ab_result_sac_hit).label("total_sac_hit"),
                func.sum(db.PitchFx.ab_result_sac_fly).label("total_sac_fly"),
                func.sum(db.PitchFx.ab_result_unclear).label("total_unclear_results"),
                func.sum(db.PitchFx.pitch_type_int.distinct()).label("pitch_type_int"),
                db.PitchFx.team_pitching_id.label("team_pitching_id"),
                db.PitchFx.team_batting_id.label("team_batting_id"),
                db.PitchFx.game_status_id.label("game_status_id"),
                db.PitchFx.date_id.label("date_id"),
                db.PitchFx.season_id.label("season_id"),
            ]
        )
        .where(
            and_(
                db.PitchFx.is_invalid_ibb == 0,
                db.PitchFx.is_out_of_sequence == 0,
                db.PitchFx.stand == "L",
                db.PitchFx.p_throws == "R",
            )
        )
        .select_from(db.PitchFx)
        .group_by(db.PitchFx.batter_id)
        .group_by(db.PitchFx.game_status_id)
        .order_by(db.PitchFx.batter_id_mlb)
        .order_by(db.PitchFx.date_id),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pfx_metrics_vs_rhp_as_lhb_for_game_for_batter(cls, db_engine, mlb_id, game_id):
        s = select([cls]).where(and_(cls.mlb_id == mlb_id, cls.bbref_game_id == game_id))
        results = db_engine.execute(s).fetchall()
        return [dict(row) for row in results][0] if results else {}


class Batter_PitchType_For_Game_vs_RHP_as_LHB_View(db.Base):
    __table__ = create_view(
        name="batter_pitchtype_vs_rhp_as_lhb_for_game",
        selectable=select(
            [
                db.PitchFx.batter_id.label("id"),
                db.PitchFx.batter_id_mlb.label("mlb_id"),
                db.PitchFx.bbref_game_id.label("bbref_game_id"),
                db.PitchFx.stand.label("bat_stand"),
                db.PitchFx.p_throws.label("p_throws"),
                db.PitchFx.mlbam_pitch_name.label("pitch_type"),
                func.count(db.PitchFx.id).label("total_pitches"),
                func.sum(db.PitchFx.is_final_pitch_of_ab).label("total_pa"),
                total_at_bats,
                func.sum(db.PitchFx.ab_result_out).label("total_outs"),
                func.sum(db.PitchFx.ab_result_hit).label("total_hits"),
                func.sum(db.PitchFx.ab_result_bb).label("total_bb"),
                func.sum(db.PitchFx.ab_result_k).label("total_k"),
                avg,
                obp,
                slg,
                ops,
                iso,
                fly_ball_rate,
                ground_ball_rate,
                line_drive_rate,
                pop_up_rate,
                bb_rate,
                k_rate,
                hr_per_fb,
                zone_rate,
                called_strike_rate,
                swinging_strike_rate,
                whiff_rate,
                csw_rate,
                o_swing_rate,
                z_swing_rate,
                swing_rate,
                o_contact_rate,
                z_contact_rate,
                contact_rate,
                func.sum(db.PitchFx.batter_did_swing).label("total_swings"),
                func.sum(db.PitchFx.batter_made_contact).label("total_swings_made_contact"),
                func.sum(db.PitchFx.called_strike).label("total_called_strikes"),
                func.sum(db.PitchFx.swinging_strike).label("total_swinging_strikes"),
                func.sum(db.PitchFx.inside_strike_zone).label("total_inside_strike_zone"),
                func.sum(db.PitchFx.outside_strike_zone).label("total_outside_strike_zone"),
                func.sum(db.PitchFx.swing_inside_zone).label("total_swings_inside_zone"),
                func.sum(db.PitchFx.swing_outside_zone).label("total_swings_outside_zone"),
                func.sum(db.PitchFx.contact_inside_zone).label("total_contact_inside_zone"),
                func.sum(db.PitchFx.contact_outside_zone).label("total_contact_outside_zone"),
                func.sum(db.PitchFx.is_batted_ball).label("total_batted_balls"),
                func.sum(db.PitchFx.is_ground_ball).label("total_ground_balls"),
                func.sum(db.PitchFx.is_line_drive).label("total_line_drives"),
                func.sum(db.PitchFx.is_fly_ball).label("total_fly_balls"),
                func.sum(db.PitchFx.is_pop_up).label("total_pop_ups"),
                func.sum(db.PitchFx.ab_result_single).label("total_singles"),
                func.sum(db.PitchFx.ab_result_double).label("total_doubles"),
                func.sum(db.PitchFx.ab_result_triple).label("total_triples"),
                func.sum(db.PitchFx.ab_result_homerun).label("total_homeruns"),
                func.sum(db.PitchFx.ab_result_ibb).label("total_ibb"),
                func.sum(db.PitchFx.ab_result_hbp).label("total_hbp"),
                func.sum(db.PitchFx.ab_result_error).label("total_errors"),
                func.sum(db.PitchFx.ab_result_sac_hit).label("total_sac_hit"),
                func.sum(db.PitchFx.ab_result_sac_fly).label("total_sac_fly"),
                func.sum(db.PitchFx.ab_result_unclear).label("total_unclear_results"),
                func.sum(db.PitchFx.pitch_type_int.distinct()).label("pitch_type_int"),
                db.PitchFx.team_pitching_id.label("team_pitching_id"),
                db.PitchFx.team_batting_id.label("team_batting_id"),
                db.PitchFx.game_status_id.label("game_status_id"),
                db.PitchFx.date_id.label("date_id"),
                db.PitchFx.season_id.label("season_id"),
            ]
        )
        .where(
            and_(
                db.PitchFx.is_invalid_ibb == 0,
                db.PitchFx.is_out_of_sequence == 0,
                db.PitchFx.stand == "L",
                db.PitchFx.p_throws == "R",
            )
        )
        .select_from(db.PitchFx)
        .group_by(db.PitchFx.batter_id)
        .group_by(db.PitchFx.game_status_id)
        .group_by(db.PitchFx.mlbam_pitch_name)
        .order_by(db.PitchFx.batter_id_mlb)
        .order_by(db.PitchFx.date_id),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pfx_metrics_vs_rhp_as_lhb_for_game_for_batter(cls, db_engine, mlb_id, game_id):
        s = select([cls]).where(and_(cls.mlb_id == mlb_id, cls.bbref_game_id == game_id))
        results = db_engine.execute(s).fetchall()
        return [dict(row) for row in results] if results else []


class Batter_PitchFx_For_Game_vs_LHP_as_RHB_View(db.Base):
    __table__ = create_view(
        name="batter_pitchfx_vs_lhp_as_rhb_for_game",
        selectable=select(
            [
                db.PitchFx.batter_id.label("id"),
                db.PitchFx.batter_id_mlb.label("mlb_id"),
                db.PitchFx.bbref_game_id.label("bbref_game_id"),
                db.PitchFx.stand.label("bat_stand"),
                db.PitchFx.p_throws.label("p_throws"),
                literal("ALL").label("pitch_type"),
                func.count(db.PitchFx.id).label("total_pitches"),
                func.sum(db.PitchFx.is_final_pitch_of_ab).label("total_pa"),
                total_at_bats,
                func.sum(db.PitchFx.ab_result_out).label("total_outs"),
                func.sum(db.PitchFx.ab_result_hit).label("total_hits"),
                func.sum(db.PitchFx.ab_result_bb).label("total_bb"),
                func.sum(db.PitchFx.ab_result_k).label("total_k"),
                avg,
                obp,
                slg,
                ops,
                iso,
                fly_ball_rate,
                ground_ball_rate,
                line_drive_rate,
                pop_up_rate,
                bb_rate,
                k_rate,
                hr_per_fb,
                zone_rate,
                called_strike_rate,
                swinging_strike_rate,
                whiff_rate,
                csw_rate,
                o_swing_rate,
                z_swing_rate,
                swing_rate,
                o_contact_rate,
                z_contact_rate,
                contact_rate,
                func.sum(db.PitchFx.batter_did_swing).label("total_swings"),
                func.sum(db.PitchFx.batter_made_contact).label("total_swings_made_contact"),
                func.sum(db.PitchFx.called_strike).label("total_called_strikes"),
                func.sum(db.PitchFx.swinging_strike).label("total_swinging_strikes"),
                func.sum(db.PitchFx.inside_strike_zone).label("total_inside_strike_zone"),
                func.sum(db.PitchFx.outside_strike_zone).label("total_outside_strike_zone"),
                func.sum(db.PitchFx.swing_inside_zone).label("total_swings_inside_zone"),
                func.sum(db.PitchFx.swing_outside_zone).label("total_swings_outside_zone"),
                func.sum(db.PitchFx.contact_inside_zone).label("total_contact_inside_zone"),
                func.sum(db.PitchFx.contact_outside_zone).label("total_contact_outside_zone"),
                func.sum(db.PitchFx.is_batted_ball).label("total_batted_balls"),
                func.sum(db.PitchFx.is_ground_ball).label("total_ground_balls"),
                func.sum(db.PitchFx.is_line_drive).label("total_line_drives"),
                func.sum(db.PitchFx.is_fly_ball).label("total_fly_balls"),
                func.sum(db.PitchFx.is_pop_up).label("total_pop_ups"),
                func.sum(db.PitchFx.ab_result_single).label("total_singles"),
                func.sum(db.PitchFx.ab_result_double).label("total_doubles"),
                func.sum(db.PitchFx.ab_result_triple).label("total_triples"),
                func.sum(db.PitchFx.ab_result_homerun).label("total_homeruns"),
                func.sum(db.PitchFx.ab_result_ibb).label("total_ibb"),
                func.sum(db.PitchFx.ab_result_hbp).label("total_hbp"),
                func.sum(db.PitchFx.ab_result_error).label("total_errors"),
                func.sum(db.PitchFx.ab_result_sac_hit).label("total_sac_hit"),
                func.sum(db.PitchFx.ab_result_sac_fly).label("total_sac_fly"),
                func.sum(db.PitchFx.ab_result_unclear).label("total_unclear_results"),
                func.sum(db.PitchFx.pitch_type_int.distinct()).label("pitch_type_int"),
                db.PitchFx.team_pitching_id.label("team_pitching_id"),
                db.PitchFx.team_batting_id.label("team_batting_id"),
                db.PitchFx.game_status_id.label("game_status_id"),
                db.PitchFx.date_id.label("date_id"),
                db.PitchFx.season_id.label("season_id"),
            ]
        )
        .where(
            and_(
                db.PitchFx.is_invalid_ibb == 0,
                db.PitchFx.is_out_of_sequence == 0,
                db.PitchFx.stand == "R",
                db.PitchFx.p_throws == "L",
            )
        )
        .select_from(db.PitchFx)
        .group_by(db.PitchFx.batter_id)
        .group_by(db.PitchFx.game_status_id)
        .order_by(db.PitchFx.batter_id_mlb)
        .order_by(db.PitchFx.date_id),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pfx_metrics_vs_lhp_as_rhb_for_game_for_batter(cls, db_engine, mlb_id, game_id):
        s = select([cls]).where(and_(cls.mlb_id == mlb_id, cls.bbref_game_id == game_id))
        results = db_engine.execute(s).fetchall()
        return [dict(row) for row in results][0] if results else {}


class Batter_PitchType_For_Game_vs_LHP_as_RHB_View(db.Base):
    __table__ = create_view(
        name="batter_pitchtype_vs_lhp_as_rhb_for_game",
        selectable=select(
            [
                db.PitchFx.batter_id.label("id"),
                db.PitchFx.batter_id_mlb.label("mlb_id"),
                db.PitchFx.bbref_game_id.label("bbref_game_id"),
                db.PitchFx.stand.label("bat_stand"),
                db.PitchFx.p_throws.label("p_throws"),
                db.PitchFx.mlbam_pitch_name.label("pitch_type"),
                func.count(db.PitchFx.id).label("total_pitches"),
                func.sum(db.PitchFx.is_final_pitch_of_ab).label("total_pa"),
                total_at_bats,
                func.sum(db.PitchFx.ab_result_out).label("total_outs"),
                func.sum(db.PitchFx.ab_result_hit).label("total_hits"),
                func.sum(db.PitchFx.ab_result_bb).label("total_bb"),
                func.sum(db.PitchFx.ab_result_k).label("total_k"),
                avg,
                obp,
                slg,
                ops,
                iso,
                fly_ball_rate,
                ground_ball_rate,
                line_drive_rate,
                pop_up_rate,
                bb_rate,
                k_rate,
                hr_per_fb,
                zone_rate,
                called_strike_rate,
                swinging_strike_rate,
                whiff_rate,
                csw_rate,
                o_swing_rate,
                z_swing_rate,
                swing_rate,
                o_contact_rate,
                z_contact_rate,
                contact_rate,
                func.sum(db.PitchFx.batter_did_swing).label("total_swings"),
                func.sum(db.PitchFx.batter_made_contact).label("total_swings_made_contact"),
                func.sum(db.PitchFx.called_strike).label("total_called_strikes"),
                func.sum(db.PitchFx.swinging_strike).label("total_swinging_strikes"),
                func.sum(db.PitchFx.inside_strike_zone).label("total_inside_strike_zone"),
                func.sum(db.PitchFx.outside_strike_zone).label("total_outside_strike_zone"),
                func.sum(db.PitchFx.swing_inside_zone).label("total_swings_inside_zone"),
                func.sum(db.PitchFx.swing_outside_zone).label("total_swings_outside_zone"),
                func.sum(db.PitchFx.contact_inside_zone).label("total_contact_inside_zone"),
                func.sum(db.PitchFx.contact_outside_zone).label("total_contact_outside_zone"),
                func.sum(db.PitchFx.is_batted_ball).label("total_batted_balls"),
                func.sum(db.PitchFx.is_ground_ball).label("total_ground_balls"),
                func.sum(db.PitchFx.is_line_drive).label("total_line_drives"),
                func.sum(db.PitchFx.is_fly_ball).label("total_fly_balls"),
                func.sum(db.PitchFx.is_pop_up).label("total_pop_ups"),
                func.sum(db.PitchFx.ab_result_single).label("total_singles"),
                func.sum(db.PitchFx.ab_result_double).label("total_doubles"),
                func.sum(db.PitchFx.ab_result_triple).label("total_triples"),
                func.sum(db.PitchFx.ab_result_homerun).label("total_homeruns"),
                func.sum(db.PitchFx.ab_result_ibb).label("total_ibb"),
                func.sum(db.PitchFx.ab_result_hbp).label("total_hbp"),
                func.sum(db.PitchFx.ab_result_error).label("total_errors"),
                func.sum(db.PitchFx.ab_result_sac_hit).label("total_sac_hit"),
                func.sum(db.PitchFx.ab_result_sac_fly).label("total_sac_fly"),
                func.sum(db.PitchFx.ab_result_unclear).label("total_unclear_results"),
                func.sum(db.PitchFx.pitch_type_int.distinct()).label("pitch_type_int"),
                db.PitchFx.team_pitching_id.label("team_pitching_id"),
                db.PitchFx.team_batting_id.label("team_batting_id"),
                db.PitchFx.game_status_id.label("game_status_id"),
                db.PitchFx.date_id.label("date_id"),
                db.PitchFx.season_id.label("season_id"),
            ]
        )
        .where(
            and_(
                db.PitchFx.is_invalid_ibb == 0,
                db.PitchFx.is_out_of_sequence == 0,
                db.PitchFx.stand == "R",
                db.PitchFx.p_throws == "L",
            )
        )
        .select_from(db.PitchFx)
        .group_by(db.PitchFx.batter_id)
        .group_by(db.PitchFx.game_status_id)
        .group_by(db.PitchFx.mlbam_pitch_name)
        .order_by(db.PitchFx.batter_id_mlb)
        .order_by(db.PitchFx.date_id),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pfx_metrics_vs_lhp_as_rhb_for_game_for_batter(cls, db_engine, mlb_id, game_id):
        s = select([cls]).where(and_(cls.mlb_id == mlb_id, cls.bbref_game_id == game_id))
        results = db_engine.execute(s).fetchall()
        return [dict(row) for row in results] if results else []


class Batter_PitchFx_For_Game_vs_LHP_as_LHB_View(db.Base):
    __table__ = create_view(
        name="batter_pitchfx_vs_lhp_as_lhb_for_game",
        selectable=select(
            [
                db.PitchFx.batter_id.label("id"),
                db.PitchFx.batter_id_mlb.label("mlb_id"),
                db.PitchFx.bbref_game_id.label("bbref_game_id"),
                db.PitchFx.stand.label("bat_stand"),
                db.PitchFx.p_throws.label("p_throws"),
                literal("ALL").label("pitch_type"),
                func.count(db.PitchFx.id).label("total_pitches"),
                func.sum(db.PitchFx.is_final_pitch_of_ab).label("total_pa"),
                total_at_bats,
                func.sum(db.PitchFx.ab_result_out).label("total_outs"),
                func.sum(db.PitchFx.ab_result_hit).label("total_hits"),
                func.sum(db.PitchFx.ab_result_bb).label("total_bb"),
                func.sum(db.PitchFx.ab_result_k).label("total_k"),
                avg,
                obp,
                slg,
                ops,
                iso,
                fly_ball_rate,
                ground_ball_rate,
                line_drive_rate,
                pop_up_rate,
                bb_rate,
                k_rate,
                hr_per_fb,
                zone_rate,
                called_strike_rate,
                swinging_strike_rate,
                whiff_rate,
                csw_rate,
                o_swing_rate,
                z_swing_rate,
                swing_rate,
                o_contact_rate,
                z_contact_rate,
                contact_rate,
                func.sum(db.PitchFx.batter_did_swing).label("total_swings"),
                func.sum(db.PitchFx.batter_made_contact).label("total_swings_made_contact"),
                func.sum(db.PitchFx.called_strike).label("total_called_strikes"),
                func.sum(db.PitchFx.swinging_strike).label("total_swinging_strikes"),
                func.sum(db.PitchFx.inside_strike_zone).label("total_inside_strike_zone"),
                func.sum(db.PitchFx.outside_strike_zone).label("total_outside_strike_zone"),
                func.sum(db.PitchFx.swing_inside_zone).label("total_swings_inside_zone"),
                func.sum(db.PitchFx.swing_outside_zone).label("total_swings_outside_zone"),
                func.sum(db.PitchFx.contact_inside_zone).label("total_contact_inside_zone"),
                func.sum(db.PitchFx.contact_outside_zone).label("total_contact_outside_zone"),
                func.sum(db.PitchFx.is_batted_ball).label("total_batted_balls"),
                func.sum(db.PitchFx.is_ground_ball).label("total_ground_balls"),
                func.sum(db.PitchFx.is_line_drive).label("total_line_drives"),
                func.sum(db.PitchFx.is_fly_ball).label("total_fly_balls"),
                func.sum(db.PitchFx.is_pop_up).label("total_pop_ups"),
                func.sum(db.PitchFx.ab_result_single).label("total_singles"),
                func.sum(db.PitchFx.ab_result_double).label("total_doubles"),
                func.sum(db.PitchFx.ab_result_triple).label("total_triples"),
                func.sum(db.PitchFx.ab_result_homerun).label("total_homeruns"),
                func.sum(db.PitchFx.ab_result_ibb).label("total_ibb"),
                func.sum(db.PitchFx.ab_result_hbp).label("total_hbp"),
                func.sum(db.PitchFx.ab_result_error).label("total_errors"),
                func.sum(db.PitchFx.ab_result_sac_hit).label("total_sac_hit"),
                func.sum(db.PitchFx.ab_result_sac_fly).label("total_sac_fly"),
                func.sum(db.PitchFx.ab_result_unclear).label("total_unclear_results"),
                func.sum(db.PitchFx.pitch_type_int.distinct()).label("pitch_type_int"),
                db.PitchFx.team_pitching_id.label("team_pitching_id"),
                db.PitchFx.team_batting_id.label("team_batting_id"),
                db.PitchFx.game_status_id.label("game_status_id"),
                db.PitchFx.date_id.label("date_id"),
                db.PitchFx.season_id.label("season_id"),
            ]
        )
        .where(
            and_(
                db.PitchFx.is_invalid_ibb == 0,
                db.PitchFx.is_out_of_sequence == 0,
                db.PitchFx.stand == "L",
                db.PitchFx.p_throws == "L",
            )
        )
        .select_from(db.PitchFx)
        .group_by(db.PitchFx.batter_id)
        .group_by(db.PitchFx.game_status_id)
        .order_by(db.PitchFx.batter_id_mlb)
        .order_by(db.PitchFx.date_id),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pfx_metrics_vs_lhp_as_lhb_for_game_for_batter(cls, db_engine, mlb_id, game_id):
        s = select([cls]).where(and_(cls.mlb_id == mlb_id, cls.bbref_game_id == game_id))
        results = db_engine.execute(s).fetchall()
        return [dict(row) for row in results][0] if results else {}


class Batter_PitchType_For_Game_vs_LHP_as_LHB_View(db.Base):
    __table__ = create_view(
        name="batter_pitchtype_vs_lhp_as_lhb_for_game",
        selectable=select(
            [
                db.PitchFx.batter_id.label("id"),
                db.PitchFx.batter_id_mlb.label("mlb_id"),
                db.PitchFx.bbref_game_id.label("bbref_game_id"),
                db.PitchFx.stand.label("bat_stand"),
                db.PitchFx.p_throws.label("p_throws"),
                db.PitchFx.mlbam_pitch_name.label("pitch_type"),
                func.count(db.PitchFx.id).label("total_pitches"),
                func.sum(db.PitchFx.is_final_pitch_of_ab).label("total_pa"),
                total_at_bats,
                func.sum(db.PitchFx.ab_result_out).label("total_outs"),
                func.sum(db.PitchFx.ab_result_hit).label("total_hits"),
                func.sum(db.PitchFx.ab_result_bb).label("total_bb"),
                func.sum(db.PitchFx.ab_result_k).label("total_k"),
                avg,
                obp,
                slg,
                ops,
                iso,
                fly_ball_rate,
                ground_ball_rate,
                line_drive_rate,
                pop_up_rate,
                bb_rate,
                k_rate,
                hr_per_fb,
                zone_rate,
                called_strike_rate,
                swinging_strike_rate,
                whiff_rate,
                csw_rate,
                o_swing_rate,
                z_swing_rate,
                swing_rate,
                o_contact_rate,
                z_contact_rate,
                contact_rate,
                func.sum(db.PitchFx.batter_did_swing).label("total_swings"),
                func.sum(db.PitchFx.batter_made_contact).label("total_swings_made_contact"),
                func.sum(db.PitchFx.called_strike).label("total_called_strikes"),
                func.sum(db.PitchFx.swinging_strike).label("total_swinging_strikes"),
                func.sum(db.PitchFx.inside_strike_zone).label("total_inside_strike_zone"),
                func.sum(db.PitchFx.outside_strike_zone).label("total_outside_strike_zone"),
                func.sum(db.PitchFx.swing_inside_zone).label("total_swings_inside_zone"),
                func.sum(db.PitchFx.swing_outside_zone).label("total_swings_outside_zone"),
                func.sum(db.PitchFx.contact_inside_zone).label("total_contact_inside_zone"),
                func.sum(db.PitchFx.contact_outside_zone).label("total_contact_outside_zone"),
                func.sum(db.PitchFx.is_batted_ball).label("total_batted_balls"),
                func.sum(db.PitchFx.is_ground_ball).label("total_ground_balls"),
                func.sum(db.PitchFx.is_line_drive).label("total_line_drives"),
                func.sum(db.PitchFx.is_fly_ball).label("total_fly_balls"),
                func.sum(db.PitchFx.is_pop_up).label("total_pop_ups"),
                func.sum(db.PitchFx.ab_result_single).label("total_singles"),
                func.sum(db.PitchFx.ab_result_double).label("total_doubles"),
                func.sum(db.PitchFx.ab_result_triple).label("total_triples"),
                func.sum(db.PitchFx.ab_result_homerun).label("total_homeruns"),
                func.sum(db.PitchFx.ab_result_ibb).label("total_ibb"),
                func.sum(db.PitchFx.ab_result_hbp).label("total_hbp"),
                func.sum(db.PitchFx.ab_result_error).label("total_errors"),
                func.sum(db.PitchFx.ab_result_sac_hit).label("total_sac_hit"),
                func.sum(db.PitchFx.ab_result_sac_fly).label("total_sac_fly"),
                func.sum(db.PitchFx.ab_result_unclear).label("total_unclear_results"),
                func.sum(db.PitchFx.pitch_type_int.distinct()).label("pitch_type_int"),
                db.PitchFx.team_pitching_id.label("team_pitching_id"),
                db.PitchFx.team_batting_id.label("team_batting_id"),
                db.PitchFx.game_status_id.label("game_status_id"),
                db.PitchFx.date_id.label("date_id"),
                db.PitchFx.season_id.label("season_id"),
            ]
        )
        .where(
            and_(
                db.PitchFx.is_invalid_ibb == 0,
                db.PitchFx.is_out_of_sequence == 0,
                db.PitchFx.stand == "L",
                db.PitchFx.p_throws == "L",
            )
        )
        .select_from(db.PitchFx)
        .group_by(db.PitchFx.batter_id)
        .group_by(db.PitchFx.game_status_id)
        .group_by(db.PitchFx.mlbam_pitch_name)
        .order_by(db.PitchFx.batter_id_mlb)
        .order_by(db.PitchFx.date_id),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )

    @classmethod
    def get_pfx_metrics_vs_lhp_as_lhb_for_game_for_batter(cls, db_engine, mlb_id, game_id):
        s = select([cls]).where(and_(cls.mlb_id == mlb_id, cls.bbref_game_id == game_id))
        results = db_engine.execute(s).fetchall()
        return [dict(row) for row in results] if results else []



import os
import subprocess

from bullet.cursor import hide as cursor_hidden
from bullet.utils import moveCursorUp

from vigorish.cli.components.viewers.page_viewer import PageViewer


class TableViewer(PageViewer):
    def __init__(
        self,
        table_list,
        prompt,
        confirm_only=False,
        yes_choice="YES",
        no_choice="NO",
        heading_color=None,
        message_color=None,
        table_color=None,
    ):
        super(TableViewer, self).__init__(
            pages=table_list,
            prompt=prompt,
            confirm_only=confirm_only,
            yes_choice=yes_choice,
            no_choice=no_choice,
            heading_color=heading_color,
            text_color=message_color,
        )
        self.table_color = table_color

    def launch(self):
        return self.interactive_mode() if os.environ.get("INTERACTIVE_MODE") == "YES" else self.script_mode()

    def interactive_mode(self):
        while True:
            subprocess.run(["clear"])
            self.current_page.display(
                heading_color=self.heading_color,
                message_color=self.text_color,
                table_color=self.table_color,
            )
            self.print_page_nav()
            self.needs_update = False
            self.print_prompt()
            if not self.confirm_only:
                self.renderBullets()
            moveCursorUp(len(self.choices) - self.pos)
            with cursor_hidden():
                while True:
                    user_selection = self.handle_input()
                    if self.needs_update:
                        break
                    if user_selection is not None:
                        return self.choices_dict.get(user_selection)

    def script_mode(self):
        subprocess.run(["clear"])
        for page in self.pages:
            page.display(
                heading_color=self.heading_color,
                message_color=self.text_color,
                table_color=self.table_color,
            )
